import asyncio
from typing import Dict, Generator, Optional

import mock
import pytest
from decoy import Decoy

from opentrons_shared_data.errors.exceptions import (
    MoveConditionNotMetError,
    PositionUnknownError,
)
from opentrons_shared_data.pipette.types import PipetteModel

from opentrons import config, types
from opentrons import hardware_control as hc
from opentrons.calibration_storage.types import (
    CalibrationStatus,
    SourceType,
)
from opentrons.config.types import AxisDict, GantryLoad
from opentrons.drivers.smoothie_drivers.simulator import SimulatingDriver
from opentrons.hardware_control import API
from opentrons.hardware_control.backends.simulator import Simulator
from opentrons.hardware_control.errors import (
    InvalidCriticalPoint,
    OutOfBoundsMove,
)
from opentrons.hardware_control.robot_calibration import (
    DeckCalibration,
    RobotCalibration,
)
from opentrons.hardware_control.thread_manager import ThreadManager
from opentrons.hardware_control.types import (
    Axis,
    CriticalPoint,
    MotionChecks,
)


async def test_controller_must_home(hardware_api: API) -> None:
    abs_position = types.Point(30, 20, 10)
    mount = types.Mount.RIGHT
    home = mock.AsyncMock()
    hardware_api.home = home  # type: ignore[method-assign]
    await hardware_api.move_to(mount, abs_position)
    home.assert_called_once()


async def test_home_specific_sim(hardware_api: API) -> None:
    await hardware_api.home()
    await hardware_api.move_to(types.Mount.RIGHT, types.Point(0, 10, 20))
    # Avoid the autoretract when moving two difference instruments
    hardware_api._last_moved_mount = None
    await hardware_api.move_rel(types.Mount.LEFT, types.Point(0, 0, -20))
    await hardware_api.home([Axis.Z, Axis.C])
    assert hardware_api._current_position == {
        Axis.X: 0,
        Axis.Y: 10,
        Axis.Z: 218,
        Axis.A: -10,
        Axis.B: 19,
        Axis.C: 19,
    }


async def test_retract(hardware_api: API) -> None:
    await hardware_api.home()
    await hardware_api.move_to(types.Mount.RIGHT, types.Point(0, 10, 20))
    await hardware_api.retract(types.Mount.RIGHT, 10)
    assert hardware_api._current_position == {
        Axis.X: 0,
        Axis.Y: 10,
        Axis.Z: 218,
        Axis.A: 218,
        Axis.B: 19,
        Axis.C: 19,
    }


@pytest.fixture
def mock_home(ot3_hardware: ThreadManager[API]) -> Generator[mock.Mock, None, None]:
    with mock.patch.object(ot3_hardware._backend, "home") as mock_home:
        mock_home.return_value = {
            Axis.X: 0,
            Axis.Y: 0,
            Axis.Z_L: 0,
            Axis.Z_R: 0,
            Axis.P_L: 0,
            Axis.P_R: 0,
            Axis.Z_G: 0,
            Axis.G: 0,
        }
        yield mock_home


async def test_home(ot3_hardware: ThreadManager[API], mock_home: mock.Mock) -> None:
    with mock.patch("opentrons.hardware_control.ot3api.deck_from_machine") as dfm_mock:
        dfm_mock.return_value = {Axis.X: 20}
        await ot3_hardware._home([Axis.X])
        assert ot3_hardware.gantry_load == GantryLoad.LOW_THROUGHPUT
        mock_home.assert_called_once_with([Axis.X], GantryLoad.LOW_THROUGHPUT)
        assert dfm_mock.call_count == 2
        dfm_mock.assert_called_with(
            machine_pos=mock_home.return_value,
            attitude=ot3_hardware._robot_calibration.deck_calibration.attitude,
            offset=ot3_hardware._robot_calibration.carriage_offset,
            robot_type="OT-3 Standard",
        )
    assert ot3_hardware._current_position[Axis.X] == 20


async def test_home_unmet(
    ot3_hardware: ThreadManager[API], mock_home: mock.Mock
) -> None:
    mock_home.side_effect = MoveConditionNotMetError()
    with pytest.raises(MoveConditionNotMetError):
        await ot3_hardware.home([Axis.X])
    assert ot3_hardware.gantry_load == GantryLoad.LOW_THROUGHPUT
    mock_home.assert_called_once_with([Axis.X], GantryLoad.LOW_THROUGHPUT)
    assert ot3_hardware._current_position == {}


async def test_move(hardware_api: API) -> None:
    abs_position = types.Point(30, 20, 10)
    mount = types.Mount.RIGHT
    target_position1 = {
        Axis.X: 30,
        Axis.Y: 20,
        Axis.Z: 218,
        Axis.A: -20,
        Axis.B: 19,
        Axis.C: 19,
    }
    await hardware_api.home()
    await hardware_api.move_to(mount, abs_position)
    assert hardware_api._current_position == target_position1

    # This assert implicitly checks that the mount offset is not applied to
    # relative moves; if you change this to move_to, the offset will be
    # applied again
    rel_position = types.Point(30, 20, -10)
    mount2 = types.Mount.LEFT
    target_position2 = {
        Axis.X: 60,
        Axis.Y: 40,
        Axis.Z: 208,
        Axis.A: 218,  # The other instrument is retracted
        Axis.B: 19,
        Axis.C: 19,
    }
    await hardware_api.move_rel(mount2, rel_position)
    assert hardware_api._current_position == target_position2


async def test_move_extras_passed_through(
    hardware_api: API, monkeypatch: pytest.MonkeyPatch
) -> None:
    mock_be_move = mock.AsyncMock()
    monkeypatch.setattr(hardware_api._backend, "move", mock_be_move)
    await hardware_api.home()
    await hardware_api.move_to(types.Mount.RIGHT, types.Point(0, 0, 0))
    assert mock_be_move.call_args_list[0][1]["speed"] is None
    assert mock_be_move.call_args_list[0][1]["axis_max_speeds"] == {}
    mock_be_move.reset_mock()
    await hardware_api.move_to(
        types.Mount.RIGHT, types.Point(1, 1, 1), speed=30, max_speeds={Axis.X: 10}
    )
    assert mock_be_move.call_args_list[0][1]["speed"] == 30
    assert mock_be_move.call_args_list[0][1]["axis_max_speeds"] == {"X": 10}
    mock_be_move.reset_mock()
    await hardware_api.move_rel(
        types.Mount.LEFT, types.Point(1, 1, 1), speed=40, max_speeds={Axis.Y: 20}
    )
    assert mock_be_move.call_args_list[0][1]["speed"] == 40
    assert mock_be_move.call_args_list[0][1]["axis_max_speeds"] == {"Y": 20}


async def test_mount_offset_applied(hardware_api: API) -> None:
    await hardware_api.home()
    abs_position = types.Point(30, 20, 10)
    mount = types.Mount.LEFT
    target_position = {
        Axis.X: 64,
        Axis.Y: 20,
        Axis.Z: -20,
        Axis.A: 218,
        Axis.B: 19,
        Axis.C: 19,
    }
    await hardware_api.move_to(mount, abs_position)
    assert hardware_api._current_position == target_position


@pytest.mark.parametrize(
    "critical_point",
    [
        CriticalPoint.GRIPPER_JAW_CENTER,
        CriticalPoint.GRIPPER_FRONT_CALIBRATION_PIN,
        CriticalPoint.GRIPPER_REAR_CALIBRATION_PIN,
    ],
)
async def test_gripper_critical_points_fail_on_pipettes(
    hardware_api: API, critical_point: CriticalPoint
) -> None:
    await hardware_api.home()
    assert isinstance(hardware_api._backend, Simulator)
    hardware_api._backend._attached_instruments = {
        types.Mount.LEFT: {"model": None, "id": None},
        types.Mount.RIGHT: {"model": PipetteModel("p10_single_v1"), "id": "testyness"},
    }
    await hardware_api.cache_instruments()
    with pytest.raises(InvalidCriticalPoint):
        await hardware_api.move_to(
            types.Mount.RIGHT, types.Point(0, 0, 0), critical_point=critical_point
        )


async def test_critical_point_applied(hardware_api: API) -> None:
    await hardware_api.home()
    assert isinstance(hardware_api._backend, Simulator)
    hardware_api._backend._attached_instruments = {
        types.Mount.LEFT: {"model": None, "id": None},
        types.Mount.RIGHT: {"model": PipetteModel("p10_single_v1"), "id": "testyness"},
    }
    await hardware_api.cache_instruments()
    # Our critical point is now the tip of the nozzle
    await hardware_api.move_to(types.Mount.RIGHT, types.Point(0, 0, 0))
    target_no_offset: Dict[Axis, float] = {
        Axis.X: 0,
        Axis.Y: 0,
        Axis.Z: 218,
        Axis.A: -12,  # from pipette-config.json nozzle offset
        Axis.B: 19,
        Axis.C: 19,
    }
    assert hardware_api._current_position == target_no_offset
    target = {Axis.X: 0.0, Axis.Y: 0.0, Axis.A: 0.0, Axis.C: 19.0}
    assert await hardware_api.current_position(types.Mount.RIGHT) == target
    p10_tip_length = 33
    # Specifiying critical point overrides as mount should not use nozzle
    # offset
    await hardware_api.move_to(
        types.Mount.RIGHT, types.Point(0, 0, 0), critical_point=CriticalPoint.MOUNT
    )
    assert hardware_api._current_position == {
        Axis.X: 0.0,
        Axis.Y: 0.0,
        Axis.Z: 218,
        Axis.A: -30,
        Axis.B: 19,
        Axis.C: 19,
    }
    assert await hardware_api.current_position(
        types.Mount.RIGHT, critical_point=CriticalPoint.MOUNT
    ) == {Axis.X: 0.0, Axis.Y: 0.0, Axis.A: 0, Axis.C: 19}
    # Specifying the critical point as nozzle should have the same behavior
    await hardware_api.move_to(
        types.Mount.RIGHT, types.Point(0, 0, 0), critical_point=CriticalPoint.NOZZLE
    )
    assert hardware_api._current_position == target_no_offset
    await hardware_api.pick_up_tip(types.Mount.RIGHT, p10_tip_length)
    # Now the current position (with offset applied) should change
    # pos_after_pickup + model_offset + critical point
    target[Axis.A] = 218 + 12 + (-1 * p10_tip_length)
    target_no_offset[Axis.C] = target[Axis.C] = 2
    assert await hardware_api.current_position(types.Mount.RIGHT) == target
    # This move should take the new critical point into account
    await hardware_api.move_to(types.Mount.RIGHT, types.Point(0, 0, 0))
    target_no_offset[Axis.A] = 21
    assert hardware_api._current_position == target_no_offset
    # But the position with offset should be back to the original
    target[Axis.A] = 0
    assert await hardware_api.current_position(types.Mount.RIGHT) == target
    # And removing the tip should move us back to the original
    await hardware_api.move_rel(types.Mount.RIGHT, types.Point(2.5, 0, 0))
    await hardware_api.drop_tip(types.Mount.RIGHT)
    await hardware_api.home_plunger(types.Mount.RIGHT)
    target[Axis.A] = 33 + hc.DROP_TIP_RELEASE_DISTANCE
    target_no_offset[Axis.X] = 2.5
    target[Axis.X] = 2.5
    assert await hardware_api.current_position(types.Mount.RIGHT) == target
    await hardware_api.move_to(types.Mount.RIGHT, types.Point(0, 0, 0))
    target[Axis.X] = 0
    target_no_offset[Axis.X] = 0
    target_no_offset[Axis.A] = -12
    target[Axis.A] = 0
    assert hardware_api._current_position == target_no_offset
    assert await hardware_api.current_position(types.Mount.RIGHT) == target


async def test_tip_pickup_routine(
    hardware_api: API, monkeypatch: pytest.MonkeyPatch
) -> None:
    _move = mock.Mock(side_effect=hardware_api._move)
    monkeypatch.setattr(hardware_api, "_move", _move)

    await hardware_api.home()
    assert isinstance(hardware_api._backend, Simulator)
    hardware_api._backend._attached_instruments = {
        types.Mount.LEFT: {"model": None, "id": None},
        types.Mount.RIGHT: {"model": PipetteModel("p10_single_v1"), "id": "testyness"},
    }
    await hardware_api.cache_instruments()
    mount = types.Mount.RIGHT

    presses = 1
    await hardware_api.tip_pickup_moves(mount, presses=presses)

    # tip pickup moves has an initial move to above the tip, then 2
    # moves for each press
    tip_motor_routine_num_moves = 2 * presses + 1

    assert len(_move.call_args_list) == tip_motor_routine_num_moves
    _move.reset_mock()

    # pick_up_tip should have the press moves + a plunger move both before and after
    await hardware_api.pick_up_tip(
        mount=mount, tip_length=40.0, presses=1, increment=None, prep_after=True
    )
    # pick_up_tip contains an additional retract
    assert len(_move.call_args_list) == tip_motor_routine_num_moves + 1


async def test_new_critical_point_applied(hardware_api: API) -> None:
    await hardware_api.home()
    assert isinstance(hardware_api._backend, Simulator)
    hardware_api._backend._attached_instruments = {
        types.Mount.LEFT: {"model": None, "id": None},
        types.Mount.RIGHT: {"model": PipetteModel("p10_single_v1"), "id": "testyness"},
    }
    await hardware_api.cache_instruments()
    # Our critical point is now the tip of the nozzle
    await hardware_api.move_to(types.Mount.RIGHT, types.Point(0, 0, 0))
    target_no_offset: Dict[Axis, float] = {
        Axis.X: 0,
        Axis.Y: 0,
        Axis.Z: 218,
        Axis.A: -12,  # from pipette-config.json model offset
        Axis.B: 19,
        Axis.C: 19,
    }
    assert hardware_api._current_position == target_no_offset
    target: Dict[Axis, float] = {Axis.X: 0, Axis.Y: 0, Axis.A: 0, Axis.C: 19}
    assert await hardware_api.current_position(types.Mount.RIGHT) == target
    p10_tip_length = 33
    # Specifiying critical point overrides as mount should not use model offset
    await hardware_api.move_to(
        types.Mount.RIGHT, types.Point(0, 0, 0), critical_point=CriticalPoint.MOUNT
    )
    assert hardware_api._current_position == {
        Axis.X: 0.0,
        Axis.Y: 0.0,
        Axis.Z: 218,
        Axis.A: -30,
        Axis.B: 19,
        Axis.C: 19,
    }
    assert await hardware_api.current_position(
        types.Mount.RIGHT, critical_point=CriticalPoint.MOUNT
    ) == {Axis.X: 0.0, Axis.Y: 0.0, Axis.A: 0, Axis.C: 19}
    # Specifying the critical point as nozzle should have the same behavior
    await hardware_api.move_to(
        types.Mount.RIGHT, types.Point(0, 0, 0), critical_point=CriticalPoint.NOZZLE
    )
    assert hardware_api._current_position == target_no_offset
    await hardware_api.pick_up_tip(types.Mount.RIGHT, p10_tip_length)
    # Now the current position (with offset applied) should change
    # pos_after_pickup + model_offset + critical point
    target[Axis.A] = 218 + (12) + (-1 * p10_tip_length)
    target_no_offset[Axis.C] = target[Axis.C] = 2
    assert await hardware_api.current_position(types.Mount.RIGHT) == target
    # This move should take the new critical point into account
    await hardware_api.move_to(types.Mount.RIGHT, types.Point(0, 0, 0))
    target_no_offset[Axis.A] = 21
    assert hardware_api._current_position == target_no_offset
    # But the position with offset should be back to the original
    target[Axis.A] = 0
    assert await hardware_api.current_position(types.Mount.RIGHT) == target
    # And removing the tip should move us back to the original
    await hardware_api.move_rel(types.Mount.RIGHT, types.Point(2.5, 0, 0))
    await hardware_api.drop_tip(types.Mount.RIGHT)
    await hardware_api.home_plunger(types.Mount.RIGHT)
    target[Axis.A] = 33 + hc.DROP_TIP_RELEASE_DISTANCE
    target_no_offset[Axis.X] = 2.5
    target[Axis.X] = 2.5
    assert await hardware_api.current_position(types.Mount.RIGHT) == target
    await hardware_api.move_to(types.Mount.RIGHT, types.Point(0, 0, 0))
    target[Axis.X] = 0
    target_no_offset[Axis.X] = 0
    target_no_offset[Axis.A] = -12
    target[Axis.A] = 0
    assert hardware_api._current_position == target_no_offset
    assert await hardware_api.current_position(types.Mount.RIGHT) == target


async def test_attitude_deck_cal_applied(monkeypatch: pytest.MonkeyPatch) -> None:
    new_gantry_cal = [[1.0047, -0.0046, 0.0], [0.0011, 1.0038, 0.0], [0.0, 0.0, 1.0]]
    called_with = None

    async def mock_move(
        position: AxisDict,
        speed: Optional[float] = None,
        home_flagged_axes: bool = True,
        axis_max_speeds: Optional[Dict[Axis, float]] = None,
    ) -> None:
        nonlocal called_with
        called_with = position

    hardware_api = await hc.API.build_hardware_simulator(
        loop=asyncio.get_running_loop()
    )
    monkeypatch.setattr(hardware_api._backend, "move", mock_move)
    deck_cal = RobotCalibration(
        deck_calibration=DeckCalibration(
            attitude=new_gantry_cal, source=SourceType.user, status=CalibrationStatus()
        )
    )
    hardware_api.set_robot_calibration(deck_cal)
    await hardware_api.home()

    await hardware_api.move_to(types.Mount.RIGHT, types.Point(0, 0, 0))

    assert called_with is not None

    assert called_with["X"] == 0.0
    assert called_with["Y"] == 0.0
    assert called_with["A"] == -30.0
    # Check that mount offset is also applied
    await hardware_api.move_to(types.Mount.LEFT, types.Point(0, 0, 0))
    assert round(called_with["X"], 2) == 34.16
    assert round(called_with["Y"], 2) == 0.04
    assert round(called_with["Z"], 2) == -30.0


async def test_other_mount_retracted(hardware_api: API) -> None:
    await hardware_api.home()
    await hardware_api.move_to(types.Mount.RIGHT, types.Point(0, 0, 0))
    assert await hardware_api.gantry_position(types.Mount.RIGHT) == types.Point(0, 0, 0)
    await hardware_api.move_to(types.Mount.LEFT, types.Point(20, 20, 0))
    assert await hardware_api.gantry_position(types.Mount.RIGHT) == types.Point(
        54, 20, 248
    )


async def test_shake_during_pick_up(
    hardware_api: API, monkeypatch: pytest.MonkeyPatch
) -> None:
    await hardware_api.home()
    assert isinstance(hardware_api._backend, Simulator)
    hardware_api._backend._attached_instruments = {
        types.Mount.LEFT: {"model": None, "id": None},
        types.Mount.RIGHT: {
            "model": PipetteModel("p1000_single_v2.0"),
            "id": "testyness",
        },
    }
    await hardware_api.cache_instruments()

    shake_tips_pick_up = mock.Mock(side_effect=hardware_api._build_pickup_shakes)
    monkeypatch.setattr(hardware_api, "_build_pickup_shakes", shake_tips_pick_up)

    move_rel = mock.Mock(side_effect=hardware_api.move_rel)
    monkeypatch.setattr(hardware_api, "move_rel", move_rel)

    # Test double shake for after pick up tips
    await hardware_api.pick_up_tip(types.Mount.RIGHT, 50)
    shake_tips_pick_up.assert_called_once()
    move_rel_calls = [
        mock.call(types.Mount.RIGHT, types.Point(-0.3, 0, 0), speed=50),
        mock.call(types.Mount.RIGHT, types.Point(0.6, 0, 0), speed=50),
        mock.call(types.Mount.RIGHT, types.Point(-0.3, 0, 0), speed=50),
        mock.call(types.Mount.RIGHT, types.Point(0, -0.3, 0), speed=50),
        mock.call(types.Mount.RIGHT, types.Point(0, 0.6, 0), speed=50),
        mock.call(types.Mount.RIGHT, types.Point(0, -0.3, 0), speed=50),
        mock.call(types.Mount.RIGHT, types.Point(0, 0, 20), speed=None),
    ]
    move_rel.assert_has_calls(move_rel_calls)


async def test_shake_during_drop(
    hardware_api: API, monkeypatch: pytest.MonkeyPatch
) -> None:
    await hardware_api.home()
    assert isinstance(hardware_api._backend, Simulator)
    hardware_api._backend._attached_instruments = {
        types.Mount.LEFT: {"model": None, "id": None},
        types.Mount.RIGHT: {
            "model": PipetteModel("p1000_single_v1.5"),
            "id": "testyness",
        },
    }
    await hardware_api.cache_instruments()
    hardware_api.add_tip(types.Mount.RIGHT, 50.0)
    hardware_api.set_current_tiprack_diameter(types.Mount.RIGHT, 2.0 * 4)

    shake_tips_drop = mock.Mock(side_effect=hardware_api._shake_off_tips_drop)
    monkeypatch.setattr(hardware_api, "_shake_off_tips_drop", shake_tips_drop)

    move_rel = mock.Mock(side_effect=hardware_api.move_rel)
    monkeypatch.setattr(hardware_api, "move_rel", move_rel)

    # Test drop tip shake with 25% of tiprack well diameter
    # between upper (2.25 mm) and lower limit (1.0 mm)
    shake_tips_drop.reset_mock()
    await hardware_api.drop_tip(types.Mount.RIGHT)
    shake_tips_drop.assert_called_once()
    move_rel_calls = [
        mock.call(types.Mount.RIGHT, types.Point(-2, 0, 0), speed=50),
        mock.call(types.Mount.RIGHT, types.Point(4, 0, 0), speed=50),
        mock.call(types.Mount.RIGHT, types.Point(-2, 0, 0), speed=50),
        mock.call(types.Mount.RIGHT, types.Point(0, 0, 20), speed=None),
    ]
    move_rel.assert_has_calls(move_rel_calls)

    # Test drop tip shake with 25% of tiprack well diameter
    # over upper (2.25 mm) limit
    hardware_api.add_tip(types.Mount.RIGHT, 20)
    hardware_api.set_current_tiprack_diameter(types.Mount.RIGHT, 2.3 * 4)
    shake_tips_drop.reset_mock()
    move_rel.reset_move()
    await hardware_api.drop_tip(types.Mount.RIGHT)
    move_rel_calls = [
        mock.call(types.Mount.RIGHT, types.Point(-2.25, 0, 0), speed=50),
        mock.call(types.Mount.RIGHT, types.Point(4.5, 0, 0), speed=50),
        mock.call(types.Mount.RIGHT, types.Point(-2.25, 0, 0), speed=50),
        mock.call(types.Mount.RIGHT, types.Point(0, 0, 20), speed=None),
    ]
    move_rel.assert_has_calls(move_rel_calls)

    # Test drop tip shake with 25% of tiprack well diameter
    # below lower (1.0 mm) limit
    hardware_api.add_tip(types.Mount.RIGHT, 50)
    hardware_api.set_current_tiprack_diameter(types.Mount.RIGHT, 0.9 * 4)
    shake_tips_drop.reset_mock()
    move_rel.reset_mock()
    await hardware_api.drop_tip(types.Mount.RIGHT)
    move_rel_calls = [
        mock.call(types.Mount.RIGHT, types.Point(-1, 0, 0), speed=50),
        mock.call(types.Mount.RIGHT, types.Point(2, 0, 0), speed=50),
        mock.call(types.Mount.RIGHT, types.Point(-1, 0, 0), speed=50),
        mock.call(types.Mount.RIGHT, types.Point(0, 0, 20), speed=None),
    ]
    move_rel.assert_has_calls(move_rel_calls)


async def test_move_rel_bounds(hardware_api: API) -> None:
    with pytest.raises(OutOfBoundsMove):
        await hardware_api.move_rel(
            types.Mount.RIGHT, types.Point(0, 0, 2000), check_bounds=MotionChecks.HIGH
        )


async def test_move_rel_homing_failures(hardware_api: API) -> None:
    await hardware_api.home()
    assert isinstance(hardware_api._backend._smoothie_driver, SimulatingDriver)
    hardware_api._backend._smoothie_driver._homed_flags = {
        "X": True,
        "Y": True,
        "Z": False,
        "A": True,
        "B": False,
        "C": False,
    }
    # If one axis being used isn't homed, we must get an exception
    with pytest.raises(PositionUnknownError):
        await hardware_api.move_rel(
            types.Mount.LEFT, types.Point(0, 0, 2000), fail_on_not_homed=True
        )

    # If an axis that _isn't_ being moved isn't homed, that's fine
    await hardware_api.move_rel(
        types.Mount.RIGHT, types.Point(0, 0, 2000), fail_on_not_homed=True
    )


async def test_current_position_homing_failures(hardware_api: API) -> None:
    await hardware_api.home()
    assert isinstance(hardware_api._backend._smoothie_driver, SimulatingDriver)
    hardware_api._backend._smoothie_driver._homed_flags = {
        "X": True,
        "Y": True,
        "Z": False,
        "A": True,
        "B": False,
        "C": True,
    }

    # If one axis being used isn't homed, we must get an exception
    with pytest.raises(PositionUnknownError):
        await hardware_api.current_position(
            mount=types.Mount.LEFT,
            fail_on_not_homed=True,
        )

    with pytest.raises(PositionUnknownError):
        await hardware_api.gantry_position(
            mount=types.Mount.LEFT,
            fail_on_not_homed=True,
        )

    # If an axis that _isn't_ being moved isn't homed, that's fine
    await hardware_api.current_position(
        mount=types.Mount.RIGHT,
        fail_on_not_homed=True,
    )

    await hardware_api.gantry_position(
        mount=types.Mount.RIGHT,
        fail_on_not_homed=True,
    )


async def test_home_z(decoy: Decoy) -> None:
    """It should home both Z axes by default."""
    loop = asyncio.get_running_loop()
    mock_config = decoy.mock(cls=config.types.RobotConfig)
    mock_backend = decoy.mock(cls=hc.Controller)

    subject = hc.API(backend=mock_backend, config=mock_config, loop=loop)  # type: ignore[abstract]

    decoy.when(mock_config.left_mount_offset).then_return(types.Point(0, 0, 100))
    decoy.when(await mock_backend.home(["Z", "A"])).then_return(
        {"X": 0, "Y": 0, "Z": 12.3, "A": 45.6, "B": 0, "C": 0}
    )

    await subject.home_z()

    left_result = (await subject.current_position(types.Mount.LEFT))[Axis.Z]
    right_result = (await subject.current_position(types.Mount.RIGHT))[Axis.A]

    assert left_result == 112.3 + subject.critical_point_for(types.Mount.LEFT, None).z
    assert right_result == 45.6 + subject.critical_point_for(types.Mount.RIGHT, None).z


async def test_home_z_one_mount(decoy: Decoy) -> None:
    """It should home a single Z mount."""
    loop = asyncio.get_running_loop()
    mock_config = decoy.mock(cls=config.types.RobotConfig)
    mock_backend = decoy.mock(cls=hc.Controller)

    subject = hc.API(backend=mock_backend, config=mock_config, loop=loop)  # type: ignore[abstract]

    decoy.when(await mock_backend.home(["A"])).then_return(
        {"X": 0, "Y": 0, "Z": 12.3, "A": 45.6, "B": 0, "C": 0}
    )

    await subject.home_z(types.Mount.RIGHT)

    right_result = (await subject.current_position(types.Mount.RIGHT))[Axis.A]

    assert right_result == 45.6 + subject.critical_point_for(types.Mount.RIGHT, None).z


async def test_home_z_both_mounts(decoy: Decoy) -> None:
    """It should home both Z axes if needed, even if an explicit mount is passed."""
    loop = asyncio.get_running_loop()
    mock_config = decoy.mock(cls=config.types.RobotConfig)
    mock_backend = decoy.mock(cls=hc.Controller)

    subject = hc.API(backend=mock_backend, config=mock_config, loop=loop)  # type: ignore[abstract]

    decoy.when(mock_config.left_mount_offset).then_return(types.Point(0, 0, 100))
    decoy.when(mock_backend.axis_bounds).then_return(
        {
            Axis.X: (0, 1),
            Axis.Y: (0, 1),
            Axis.Z: (0, 1),
            Axis.A: (0, 1),
        }
    )

    decoy.when(await mock_backend.home(["X", "Y", "Z", "A"])).then_return(
        {"X": 0, "Y": 0, "Z": 0, "A": 0, "B": 0, "C": 0}
    )
    decoy.when(await mock_backend.home(["Z", "A"])).then_return(
        {"X": 0, "Y": 0, "Z": 12.3, "A": 45.6, "B": 0, "C": 0}
    )

    await subject.move_rel(mount=types.Mount.LEFT, delta=types.Point(0, 0, 0))
    await subject.home_z(types.Mount.RIGHT)

    left_result = (await subject.current_position(types.Mount.LEFT))[Axis.Z]
    right_result = (await subject.current_position(types.Mount.RIGHT))[Axis.A]

    assert left_result == 112.3 + subject.critical_point_for(types.Mount.LEFT, None).z
    assert right_result == 45.6 + subject.critical_point_for(types.Mount.RIGHT, None).z


async def test_home_z_ignore_other_mount(decoy: Decoy) -> None:
    """It should only home one mount, even if other needs homing, if not `allow_other_mount`."""
    loop = asyncio.get_running_loop()
    mock_config = decoy.mock(cls=config.types.RobotConfig)
    mock_backend = decoy.mock(cls=hc.Controller)

    subject = hc.API(backend=mock_backend, config=mock_config, loop=loop)  # type: ignore[abstract]

    decoy.when(mock_config.left_mount_offset).then_return(types.Point(0, 0, 100))
    decoy.when(mock_backend.axis_bounds).then_return(
        {
            Axis.X: (0, 1),
            Axis.Y: (0, 1),
            Axis.Z: (0, 1),
            Axis.A: (0, 1),
        }
    )

    decoy.when(await mock_backend.home(["X", "Y", "Z", "A"])).then_return(
        {"X": 0, "Y": 0, "Z": 0, "A": 0, "B": 0, "C": 0}
    )
    decoy.when(await mock_backend.home(["A"])).then_return(
        {"X": 0, "Y": 0, "Z": 0, "A": 45.6, "B": 0, "C": 0}
    )

    await subject.move_rel(mount=types.Mount.LEFT, delta=types.Point(0, 0, 0))
    await subject.home_z(types.Mount.RIGHT, allow_home_other=False)

    left_result = (await subject.current_position(types.Mount.LEFT))[Axis.Z]
    right_result = (await subject.current_position(types.Mount.RIGHT))[Axis.A]

    assert left_result == 100.0 + subject.critical_point_for(types.Mount.LEFT, None).z
    assert right_result == 45.6 + subject.critical_point_for(types.Mount.RIGHT, None).z



"""Run control side-effect handler."""

from time import monotonic as time_monotonic

import pytest
from decoy import Decoy, matchers

from opentrons.protocol_engine.actions import ActionDispatcher, PauseAction, PauseSource
from opentrons.protocol_engine.execution.run_control import RunControlHandler
from opentrons.protocol_engine.state.config import Config
from opentrons.protocol_engine.state.state import StateStore
from opentrons.protocol_engine.types import DeckType


def _make_config(ignore_pause: bool) -> Config:
    return Config(
        ignore_pause=ignore_pause,
        # Robot and deck type are arbitrary.
        robot_type="OT-2 Standard",
        deck_type=DeckType.OT2_STANDARD,
    )


@pytest.fixture
def mock_state_store(decoy: Decoy) -> StateStore:
    """Get a mocked out StateStore."""
    return decoy.mock(cls=StateStore)


@pytest.fixture
def mock_action_dispatcher(decoy: Decoy) -> ActionDispatcher:
    """Get a mocked out ActionDispatcher."""
    return decoy.mock(cls=ActionDispatcher)


@pytest.fixture
def subject(
    mock_state_store: StateStore,
    mock_action_dispatcher: ActionDispatcher,
) -> RunControlHandler:
    """Create a RunControlHandler with its dependencies mocked out."""
    return RunControlHandler(
        state_store=mock_state_store,
        action_dispatcher=mock_action_dispatcher,
    )


async def test_pause(
    decoy: Decoy,
    mock_state_store: StateStore,
    mock_action_dispatcher: ActionDispatcher,
    subject: RunControlHandler,
) -> None:
    """It should be able to execute a pause."""
    decoy.when(mock_state_store.config).then_return(_make_config(ignore_pause=False))
    await subject.wait_for_resume()
    decoy.verify(
        mock_action_dispatcher.dispatch(PauseAction(source=PauseSource.PROTOCOL)),
        await mock_state_store.wait_for(
            condition=mock_state_store.commands.get_is_running
        ),
    )


async def test_pause_analysis(
    decoy: Decoy,
    mock_state_store: StateStore,
    mock_action_dispatcher: ActionDispatcher,
    subject: RunControlHandler,
) -> None:
    """It should no op during a protocol analysis."""
    decoy.when(mock_state_store.config).then_return(_make_config(ignore_pause=True))
    await subject.wait_for_resume()
    decoy.verify(
        mock_action_dispatcher.dispatch(PauseAction(source=matchers.Anything())),
        times=0,
    )


async def test_wait_for_duration(
    decoy: Decoy,
    mock_state_store: StateStore,
    subject: RunControlHandler,
) -> None:
    """It should wait for a specified duration.

    This test mixes mocks and actual functionality.
    An implementation that is "more testabe" probably involves
    re-implementing `asyncio.sleep`, which just isn't worth it.
    """
    decoy.when(mock_state_store.config).then_return(_make_config(ignore_pause=False))
    start = time_monotonic()
    await subject.wait_for_duration(seconds=0.2)
    end = time_monotonic()

    # NOTE: margin of error selected empirically
    # this is flakey test risk in CI
    assert end - start >= 0.1


async def test_wait_for_duration_ignore_pause(
    decoy: Decoy,
    mock_state_store: StateStore,
    subject: RunControlHandler,
) -> None:
    """It should wait for a specified duration.

    This test mixes mocks and actual functionality.
    An implementation that is "more testabe" probably involves
    re-implementing `asyncio.sleep`, which just isn't worth it.
    """
    decoy.when(mock_state_store.config).then_return(_make_config(ignore_pause=True))
    start = time_monotonic()
    # This wait time would be disruptively long for the test suite,
    # but it only matters when the subject has a bug and this test fails,
    # which should be rare.
    await subject.wait_for_duration(seconds=1.0)
    end = time_monotonic()

    # NOTE: margin of error selected empirically
    # this is flakey test risk in CI
    assert end - start <= 0.1



"""WaitForTasks command request, result, and implementation models."""

from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Type

from pydantic import BaseModel, Field
from typing_extensions import Literal

from ..errors.error_occurrence import ErrorOccurrence, ProtocolCommandFailedError
from ..errors.exceptions import TaskFailedError
from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData

if TYPE_CHECKING:
    from ..execution import RunControlHandler, TaskHandler
    from ..state.state import StateView


WaitForTasksCommandType = Literal["waitForTasks"]


class WaitForTasksParams(BaseModel):
    """Payload required to annotate execution with a WaitForTasks."""

    task_ids: list[str] = Field(
        ...,
        description="The list of task ids to wait for.",
    )


class WaitForTasksResult(BaseModel):
    """Result data from the execution of a WaitForTasks command."""

    task_ids: list[str] = Field(
        ...,
        description="The list of completed task ids.",
    )


class WaitForTasksImplementation(
    AbstractCommandImpl[WaitForTasksParams, SuccessData[WaitForTasksResult]]
):
    """WaitForTasks command implementation."""

    def __init__(
        self,
        task_handler: TaskHandler,
        run_control: RunControlHandler,
        state_view: StateView,
        **kwargs: object,
    ) -> None:
        self._task_handler = task_handler
        self._run_control = run_control
        self._state_view = state_view

    async def execute(
        self, params: WaitForTasksParams
    ) -> SuccessData[WaitForTasksResult]:
        """Checks for existance of task id and then asynchronously waits for the valid, specified tasks to finish."""
        # Raises the exception if we don't have a valid task id.
        for task_id in params.task_ids:
            _ = self._state_view.tasks.get(task_id)

        await self._run_control.wait_for_tasks(params.task_ids)

        failed_tasks = self._state_view.tasks.get_failed_tasks(params.task_ids)
        if failed_tasks:
            raise TaskFailedError(
                message=f"{len(failed_tasks)} tasks failed.",
                details={"failed_task_ids": failed_tasks},
                wrapping=[
                    ProtocolCommandFailedError(
                        original_error=self._state_view.tasks.get_finished(
                            task_id
                        ).error
                    )
                    for task_id in failed_tasks
                ],
            )
        return SuccessData(public=WaitForTasksResult(task_ids=params.task_ids))


class WaitForTasks(
    BaseCommand[WaitForTasksParams, WaitForTasksResult, ErrorOccurrence]
):
    """WaitForTasks command model."""

    commandType: WaitForTasksCommandType = "waitForTasks"
    params: WaitForTasksParams
    result: Optional[WaitForTasksResult] = None

    _ImplementationCls: Type[WaitForTasksImplementation] = WaitForTasksImplementation


class WaitForTasksCreate(BaseCommandCreate[WaitForTasksParams]):
    """WaitForTasks command request model."""

    commandType: WaitForTasksCommandType = "waitForTasks"
    params: WaitForTasksParams

    _CommandCls: Type[WaitForTasks] = WaitForTasks



"""CreateTimer command request, result, and implementation models."""

from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Type

from pydantic import BaseModel, Field
from typing_extensions import Literal

from ..errors.error_occurrence import ErrorOccurrence
from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData

if TYPE_CHECKING:
    from ..execution import RunControlHandler, TaskHandler


CreateTimerCommandType = Literal["createTimer"]


class CreateTimerParams(BaseModel):
    """Payload required to annotate execution with a CreateTimer."""

    time: float = Field(
        ...,
        description="The time before the timer should elapse in seconds. This is the minimum time before the timer elapses; it may in practice take longer than this.",
    )
    task_id: str | None = Field(
        None,
        description="The id of the timer task",
    )


class CreateTimerResult(BaseModel):
    """Result data from the execution of a CreateTimer command."""

    task_id: str = Field(..., description="The id of the timer task")
    time: float = Field(..., description="The same time as the parameter.")


class CreateTimerImplementation(
    AbstractCommandImpl[CreateTimerParams, SuccessData[CreateTimerResult]]
):
    """CreateTimer command implementation."""

    def __init__(
        self,
        task_handler: TaskHandler,
        run_control: RunControlHandler,
        **kwargs: object,
    ) -> None:
        self._task_handler = task_handler
        self._run_control = run_control

    async def execute(
        self, params: CreateTimerParams
    ) -> SuccessData[CreateTimerResult]:
        """No operation taken other than capturing message in command."""

        async def timer(task_handler: TaskHandler) -> None:
            async with task_handler.synchronize_concurrent("createTimer"):
                await self._run_control.wait_for_duration(params.time)

        task = await self._task_handler.create_task(timer, params.task_id)
        return SuccessData(
            public=CreateTimerResult(task_id=task.id, time=params.time),
        )


class CreateTimer(BaseCommand[CreateTimerParams, CreateTimerResult, ErrorOccurrence]):
    """CreateTimer command model."""

    commandType: CreateTimerCommandType = "createTimer"
    params: CreateTimerParams
    result: Optional[CreateTimerResult] = None

    _ImplementationCls: Type[CreateTimerImplementation] = CreateTimerImplementation


class CreateTimerCreate(BaseCommandCreate[CreateTimerParams]):
    """CreateTimer command request model."""

    commandType: CreateTimerCommandType = "createTimer"
    params: CreateTimerParams

    _CommandCls: Type[CreateTimer] = CreateTimer



import asyncio
from typing import Any, AsyncGenerator

import pytest
from decoy import Decoy

from opentrons.drivers.asyncio.communication.errors import ErrorResponse, UnhandledGcode
from opentrons.drivers.heater_shaker.simulator import SimulatingDriver
from opentrons.drivers.rpi_drivers.types import USBPort
from opentrons.drivers.types import RPM, HeaterShakerLabwareLatchStatus, Temperature
from opentrons.hardware_control import ExecutionManager, modules, poller
from opentrons.hardware_control.modules.heater_shaker import HeaterShakerReader
from opentrons.hardware_control.modules.types import (
    HeaterShakerStatus,
    ModuleDisconnectedCallback,
    ModuleErrorCallback,
    SpeedStatus,
    TemperatureStatus,
)


@pytest.fixture
def usb_port() -> USBPort:
    """A token USB port."""
    return USBPort(
        name="",
        port_number=0,
        device_path="/dev/ot_module_sim_heatershaker0",
    )


@pytest.fixture
async def simulating_module(
    usb_port: USBPort,
    mock_execution_manager: ExecutionManager,
    module_error_callback: ModuleErrorCallback,
    module_disconnected_callback: ModuleDisconnectedCallback,
) -> AsyncGenerator[modules.AbstractModule, None]:
    """Get a mocked simulating subject."""
    module = await modules.build(
        port=usb_port.device_path,
        usb_port=usb_port,
        type=modules.ModuleType["HEATER_SHAKER"],
        simulating=True,
        hw_control_loop=asyncio.get_running_loop(),
        execution_manager=mock_execution_manager,
        error_callback=module_error_callback,
        disconnected_callback=module_disconnected_callback,
    )
    assert isinstance(module, modules.AbstractModule)
    try:
        yield module
    finally:
        await module.cleanup()


@pytest.fixture
def mock_driver(decoy: Decoy) -> SimulatingDriver:
    """Get a mocked simulating driver."""
    return decoy.mock(cls=SimulatingDriver)


@pytest.fixture
async def reader_mocked_driver(mock_driver: SimulatingDriver) -> HeaterShakerReader:
    """A reader with a mocked driver."""
    return HeaterShakerReader(driver=mock_driver)


@pytest.fixture
async def simulating_module_driver_patched(
    mock_driver: SimulatingDriver,
    usb_port: USBPort,
    mock_execution_manager: ExecutionManager,
    module_error_callback: ModuleErrorCallback,
    module_disconnected_callback: ModuleDisconnectedCallback,
    reader_mocked_driver: HeaterShakerReader,
) -> AsyncGenerator[modules.AbstractModule, None]:
    """Get a mocked module with a patched driver."""
    poller_obj = poller.Poller(reader=reader_mocked_driver, interval=0.01)
    module = modules.HeaterShaker(
        port="/dev/ot_module_sim_heatershaker0",
        usb_port=usb_port,
        hw_control_loop=asyncio.get_running_loop(),
        driver=mock_driver,
        reader=reader_mocked_driver,
        poller=poller_obj,
        device_info={},
        execution_manager=mock_execution_manager,
        error_callback=module_error_callback,
        disconnected_callback=module_disconnected_callback,
    )
    await poller_obj.start()
    yield module
    await module.cleanup()


async def test_sim_state(simulating_module: modules.HeaterShaker) -> None:
    """It should forward simulated state."""
    assert simulating_module.temperature == 23
    assert simulating_module.speed == 0
    assert simulating_module.target_temperature is None
    assert simulating_module.target_speed is None
    assert simulating_module.temperature_status == TemperatureStatus.IDLE
    assert simulating_module.speed_status == SpeedStatus.IDLE
    assert simulating_module.status == HeaterShakerStatus.IDLE
    status = simulating_module.device_info
    assert status["serial"] == "dummySerialHS"
    # return v1 if sim_model is not passed
    assert status["model"] == "dummyModelHS"
    assert status["version"] == "dummyVersionHS"


async def test_sim_update(simulating_module: modules.HeaterShaker) -> None:
    """It should update its simulated state."""
    await simulating_module.start_set_temperature(10)
    await simulating_module.await_temperature(10)
    assert simulating_module.temperature == 10
    assert simulating_module.target_temperature == 10
    assert simulating_module.temperature_status == TemperatureStatus.HOLDING
    assert simulating_module.status == HeaterShakerStatus.RUNNING

    await simulating_module.set_speed(2000)
    assert simulating_module.speed == 2000
    assert simulating_module.target_speed == 2000
    assert simulating_module.speed_status == SpeedStatus.HOLDING
    assert simulating_module.status == HeaterShakerStatus.RUNNING

    await simulating_module.deactivate()
    assert simulating_module.temperature == 23
    assert simulating_module.speed == 0
    assert simulating_module.target_temperature is None
    assert simulating_module.target_speed is None
    assert simulating_module.temperature_status == TemperatureStatus.IDLE
    assert simulating_module.speed_status == SpeedStatus.IDLE


async def test_await_both(simulating_module: modules.HeaterShaker) -> None:
    """It should wait for speed and temp."""
    await simulating_module.start_set_temperature(10)
    await simulating_module.set_speed(2000)
    await simulating_module.await_temperature(10)
    assert simulating_module.temperature_status == TemperatureStatus.HOLDING
    assert simulating_module.speed_status == SpeedStatus.HOLDING


async def test_labware_latch(simulating_module: modules.HeaterShaker) -> None:
    """It should handle the latch."""
    await simulating_module.open_labware_latch()
    assert (
        await simulating_module._driver.get_labware_latch_status()
        == HeaterShakerLabwareLatchStatus.IDLE_OPEN
    )
    await simulating_module.close_labware_latch()
    assert (
        await simulating_module._driver.get_labware_latch_status()
        == HeaterShakerLabwareLatchStatus.IDLE_CLOSED
    )


async def test_initial_live_data(simulating_module: modules.HeaterShaker) -> None:
    """Should return the simulating module's initial live data."""
    assert simulating_module.live_data == {
        "data": {
            "labwareLatchStatus": "idle_unknown",
            "speedStatus": "idle",
            "temperatureStatus": "idle",
            "currentSpeed": 0,
            "currentTemp": 23,
            "targetSpeed": None,
            "targetTemp": None,
            "errorDetails": None,
        },
        "status": "idle",
    }


async def test_updated_live_data(simulating_module: modules.HeaterShaker) -> None:
    """Should update live data after module commands."""
    await simulating_module.close_labware_latch()
    await simulating_module.start_set_temperature(50)
    await simulating_module.set_speed(100)
    assert simulating_module.live_data == {
        "data": {
            "labwareLatchStatus": "idle_closed",
            "speedStatus": "holding at target",
            "temperatureStatus": "holding at target",
            "currentSpeed": 100,
            "currentTemp": 50,
            "targetSpeed": 100,
            "targetTemp": 50,
            "errorDetails": None,
        },
        "status": "running",
    }


async def test_deactivated_updated_live_data(
    simulating_module: modules.HeaterShaker,
) -> None:
    """Should update live data after module commands."""
    await simulating_module.close_labware_latch()
    await simulating_module.start_set_temperature(50)
    await simulating_module.set_speed(100)
    assert simulating_module.live_data == {
        "data": {
            "labwareLatchStatus": "idle_closed",
            "speedStatus": "holding at target",
            "temperatureStatus": "holding at target",
            "currentSpeed": 100,
            "currentTemp": 50,
            "targetSpeed": 100,
            "targetTemp": 50,
            "errorDetails": None,
        },
        "status": "running",
    }
    await simulating_module.deactivate()
    assert simulating_module.live_data == {
        "data": {
            "labwareLatchStatus": "idle_closed",
            "speedStatus": "idle",
            "temperatureStatus": "idle",
            "currentSpeed": 0,
            "currentTemp": 23,
            "targetSpeed": None,
            "targetTemp": None,
            "errorDetails": None,
        },
        "status": "idle",
    }


async def fake_get_rpm(*args: Any, **kwargs: Any) -> RPM:
    return RPM(current=500, target=500)


async def fake_get_temperature(*args: Any, **kwargs: Any) -> Temperature:
    return Temperature(current=50, target=50)


async def fake_get_latch_status(
    *args: Any, **kwargs: Any
) -> HeaterShakerLabwareLatchStatus:
    return HeaterShakerLabwareLatchStatus.IDLE_OPEN


async def test_async_error_response(
    simulating_module_driver_patched: modules.HeaterShaker,
    module_error_callback: ModuleErrorCallback,
    mock_driver: SimulatingDriver,
    decoy: Decoy,
) -> None:
    """Test that asynchronous error is detected by poller and module live data and status are updated."""
    exc = Exception("Oh no, an asynchronous error!")
    decoy.when(await mock_driver.get_error_state()).then_return(None)  # type: ignore[func-returns-value]
    decoy.when(await mock_driver.get_rpm()).then_return(RPM(current=500, target=500))
    decoy.when(await mock_driver.get_labware_latch_status()).then_return(
        HeaterShakerLabwareLatchStatus.IDLE_OPEN
    )
    decoy.when(await mock_driver.get_temperature()).then_return(
        Temperature(current=50, target=50)
    )
    await simulating_module_driver_patched._poller.wait_next_poll()
    decoy.when(await mock_driver.get_temperature()).then_raise(exc)
    with pytest.raises(Exception):
        await simulating_module_driver_patched._poller.wait_next_poll()
    decoy.verify(
        module_error_callback(
            exc, "heaterShakerModuleV1", "/dev/ot_module_sim_heatershaker0", None
        )
    )

    assert (
        simulating_module_driver_patched.live_data["data"]["errorDetails"]  # type: ignore[index,typeddict-item]
        == "Oh no, an asynchronous error!"
    )
    assert simulating_module_driver_patched.status == HeaterShakerStatus.ERROR
    decoy.reset()
    decoy.when(await mock_driver.get_error_state()).then_return(None)  # type: ignore[func-returns-value]
    decoy.when(await mock_driver.get_temperature()).then_return(
        Temperature(current=50, target=50)
    )
    decoy.when(await mock_driver.get_rpm()).then_return(RPM(current=500, target=500))
    decoy.when(await mock_driver.get_labware_latch_status()).then_return(
        HeaterShakerLabwareLatchStatus.IDLE_OPEN
    )
    await simulating_module_driver_patched._poller.wait_next_poll()
    assert simulating_module_driver_patched.live_data["data"]["errorDetails"] is None  # type: ignore[index,typeddict-item]
    assert simulating_module_driver_patched.status == HeaterShakerStatus.RUNNING  # type: ignore[comparison-overlap]


async def test_reader_ignores_get_error_state_not_available(
    mock_driver: SimulatingDriver,
    reader_mocked_driver: HeaterShakerReader,
    decoy: Decoy,
) -> None:
    """It should not raise if the module does not support get-error-state."""
    err = UnhandledGcode(
        "/dev/ot_module_sim_heatershaker0", "ERR:001:unhandled gcode", "M411"
    )
    decoy.when(await mock_driver.get_error_state()).then_raise(err)  # type: ignore[func-returns-value]
    await reader_mocked_driver.read()


async def test_reader_raises_error_from_get_error(
    mock_driver: SimulatingDriver,
    module_error_callback: ModuleErrorCallback,
    simulating_module_driver_patched: modules.HeaterShaker,
    decoy: Decoy,
) -> None:
    """It should put the error all the way out to the error callback from the reader."""
    error_state_response = ErrorResponse(
        "/dev/ot_module_sim_heatershaker0", "ERR:666:you know what it is", "M411"
    )
    decoy.when(await mock_driver.get_error_state()).then_raise(error_state_response)  # type: ignore[func-returns-value]
    with pytest.raises(ErrorResponse):
        await simulating_module_driver_patched._poller.wait_next_poll()
    decoy.verify(
        module_error_callback(
            error_state_response,
            model="heaterShakerModuleV1",
            port="/dev/ot_module_sim_heatershaker0",
            serial=None,
        )
    )



from typing import Dict

import pytest
from decoy import Decoy

from opentrons_shared_data.labware.types import LabwareDefinition2, WellDefinition2

from opentrons.calibration_storage import helpers
from opentrons.hardware_control.modules.types import (
    HeaterShakerModuleModel,
    MagneticModuleModel,
    ModuleModel,
    TemperatureModuleModel,
    ThermocyclerModuleModel,
)
from opentrons.protocol_api import Labware, labware, validation
from opentrons.protocol_api.core.labware import AbstractLabware
from opentrons.protocol_api.core.legacy import module_geometry
from opentrons.protocol_api.core.legacy.legacy_labware_core import LegacyLabwareCore
from opentrons.protocol_api.core.legacy.legacy_well_core import LegacyWellCore
from opentrons.protocol_api.core.legacy.well_geometry import WellGeometry
from opentrons.protocol_api.core.well import AbstractWellCore
from opentrons.protocols.api_support.types import APIVersion
from opentrons.types import Location, Point

test_data: Dict[str, WellDefinition2] = {
    "circular_well_json": {
        "shape": "circular",
        "depth": 40,
        "totalLiquidVolume": 100,
        "diameter": 30,
        "x": 40,
        "y": 50,
        "z": 3,
    },
    "rectangular_well_json": {
        "shape": "rectangular",
        "depth": 20,
        "totalLiquidVolume": 200,
        "xDimension": 120,
        "yDimension": 50,
        "x": 45,
        "y": 10,
        "z": 22,
    },
}


def test_well_init() -> None:
    slot = Location(Point(1, 2, 3), 1)  # type: ignore[arg-type]
    well_name = "circular_well_json"
    has_tip = False
    well1 = labware.Well(
        parent=None,  # type: ignore[arg-type]
        core=LegacyWellCore(
            well_geometry=WellGeometry(
                well_props=test_data[well_name],
                parent_point=slot.point,
                parent_object=slot.labware,  # type: ignore[arg-type]
            ),
            display_name=well_name,
            has_tip=has_tip,
            name="A1",
        ),
        api_version=APIVersion(2, 13),
    )
    assert well1.geometry.diameter == test_data[well_name]["diameter"]  # type: ignore[typeddict-item]
    assert well1.geometry._length is None
    assert well1.geometry._width is None

    well2_name = "rectangular_well_json"
    well2 = labware.Well(
        parent=None,  # type: ignore[arg-type]
        core=LegacyWellCore(
            well_geometry=WellGeometry(
                well_props=test_data[well2_name],
                parent_point=slot.point,
                parent_object=slot.labware,  # type: ignore[arg-type]
            ),
            display_name=well2_name,
            has_tip=has_tip,
            name="A1",
        ),
        api_version=APIVersion(2, 13),
    )
    assert well2.geometry.diameter is None
    assert well2.geometry._length == test_data[well2_name]["xDimension"]  # type: ignore[typeddict-item]
    assert well2.geometry._width == test_data[well2_name]["yDimension"]  # type: ignore[typeddict-item]


def test_top() -> None:
    slot = Location(Point(4, 5, 6), 1)  # type: ignore[arg-type]
    well_name = "circular_well_json"
    has_tip = False
    well = labware.Well(
        parent=None,  # type: ignore[arg-type]
        core=LegacyWellCore(
            well_geometry=WellGeometry(
                well_props=test_data[well_name],
                parent_point=slot.point,
                parent_object=slot.labware,  # type: ignore[arg-type]
            ),
            display_name=well_name,
            has_tip=has_tip,
            name="A1",
        ),
        api_version=APIVersion(2, 13),
    )
    well_data = test_data[well_name]
    expected_x = well_data["x"] + slot.point.x
    expected_y = well_data["y"] + slot.point.y
    expected_z = well_data["z"] + well_data["depth"] + slot.point.z
    assert well.top() == Location(Point(expected_x, expected_y, expected_z), well)


def test_bottom() -> None:
    slot = Location(Point(7, 8, 9), 1)  # type: ignore[arg-type]
    well_name = "rectangular_well_json"
    has_tip = False
    well = labware.Well(
        parent=None,  # type: ignore[arg-type]
        core=LegacyWellCore(
            well_geometry=WellGeometry(
                well_props=test_data[well_name],
                parent_point=slot.point,
                parent_object=slot.labware,  # type: ignore[arg-type]
            ),
            display_name=well_name,
            has_tip=has_tip,
            name="A1",
        ),
        api_version=APIVersion(2, 13),
    )
    well_data = test_data[well_name]
    expected_x = well_data["x"] + slot.point.x
    expected_y = well_data["y"] + slot.point.y
    expected_z = well_data["z"] + slot.point.z
    assert well.bottom() == Location(Point(expected_x, expected_y, expected_z), well)


def test_from_center_cartesian() -> None:
    slot1 = Location(Point(10, 11, 12), 1)  # type: ignore[arg-type]
    well_name = "circular_well_json"
    has_tip = False
    well1 = labware.Well(
        parent=None,  # type: ignore[arg-type]
        core=LegacyWellCore(
            well_geometry=WellGeometry(
                well_props=test_data[well_name],
                parent_point=slot1.point,
                parent_object=slot1.labware,  # type: ignore[arg-type]
            ),
            display_name=well_name,
            has_tip=has_tip,
            name="A1",
        ),
        api_version=APIVersion(2, 13),
    )

    percent1_x = 1
    percent1_y = 1
    percent1_z = -0.5
    point1 = well1.from_center_cartesian(percent1_x, percent1_y, percent1_z)

    # slot.x + well.x + 1 * well.diamter/2
    expected_x: float = 10 + 40 + 15
    # slot.y + well.y + 1 * well.diamter/2
    expected_y: float = 11 + 50 + 15
    # slot.z + well.z + (1 - 0.5) * well.depth/2
    expected_z: float = 12 + 3 + 20 - 10

    assert point1.x == expected_x
    assert point1.y == expected_y
    assert point1.z == expected_z

    slot2 = Location(Point(13, 14, 15), 1)  # type: ignore[arg-type]
    well2_name = "rectangular_well_json"
    has_tip = False
    well2 = labware.Well(
        parent=None,  # type: ignore[arg-type]
        core=LegacyWellCore(
            well_geometry=WellGeometry(
                well_props=test_data[well2_name],
                parent_point=slot2.point,
                parent_object=slot2.labware,  # type: ignore[arg-type]
            ),
            display_name=well2_name,
            has_tip=has_tip,
            name="A1",
        ),
        api_version=APIVersion(2, 13),
    )
    percent2_x = -0.25
    percent2_y = 0.1
    percent2_z = 0.9
    point2 = well2.geometry.from_center_cartesian(percent2_x, percent2_y, percent2_z)

    # slot.x + well.x - 0.25 * well.length/2
    expected_x = 13 + 45 - 15
    # slot.y + well.y + 0.1 * well.width/2
    expected_y = 14 + 10 + 2.5
    # slot.z + well.z + (1 + 0.9) * well.depth/2
    expected_z = 15 + 22 + 19

    assert point2.x == expected_x
    assert point2.y == expected_y
    assert point2.z == expected_z


@pytest.fixture
def corning_96_wellplate_360ul_flat_def() -> LabwareDefinition2:
    labware_name = "corning_96_wellplate_360ul_flat"
    result = labware.get_labware_definition(labware_name)
    assert result["schemaVersion"] == 2
    return result


@pytest.fixture
def corning_96_wellplate_360ul_flat(
    corning_96_wellplate_360ul_flat_def: LabwareDefinition2,
) -> Labware:
    return labware.Labware(
        core=LegacyLabwareCore(
            definition=corning_96_wellplate_360ul_flat_def,
            parent=Location(Point(0, 0, 0), "Test Slot"),
        ),
        api_version=APIVersion(2, 13),
        protocol_core=None,  # type: ignore[arg-type]
        core_map=None,  # type: ignore[arg-type]
    )


@pytest.fixture
def opentrons_96_tiprack_300ul_def() -> LabwareDefinition2:
    labware_name = "opentrons_96_tiprack_300ul"
    result = labware.get_labware_definition(labware_name)
    assert result["schemaVersion"] == 2
    return result


@pytest.fixture
def opentrons_96_tiprack_300ul(
    opentrons_96_tiprack_300ul_def: LabwareDefinition2,
) -> Labware:
    return labware.Labware(
        core=LegacyLabwareCore(
            definition=opentrons_96_tiprack_300ul_def,
            parent=Location(Point(0, 0, 0), "Test Slot"),
        ),
        api_version=APIVersion(2, 13),
        protocol_core=None,  # type: ignore[arg-type]
        core_map=None,  # type: ignore[arg-type]
    )


def test_back_compat(corning_96_wellplate_360ul_flat: Labware) -> None:
    lw = corning_96_wellplate_360ul_flat

    # Note that this test uses the display name of wells to test for equality,
    # because dimensional parameters could be subject to modification through
    # calibration, whereas here we are testing for "identity" in a way that is
    # related to the combination of well name, labware name, and slot name
    well_a1_name = repr(lw.wells_by_name()["A1"])
    well_b2_name = repr(lw.wells_by_name()["B2"])
    well_c3_name = repr(lw.wells_by_name()["C3"])

    w2 = lw.well(0)
    assert repr(w2) == well_a1_name

    w3 = lw.well("A1")
    assert repr(w3) == well_a1_name

    w4 = lw.wells("B2")
    assert repr(w4[0]) == well_b2_name

    w5 = lw.wells(9, 21, 25, 27)
    assert len(w5) == 4
    assert repr(w5[0]) == well_b2_name

    w6 = lw.wells("A1", "B2", "C3")
    assert all(
        [
            repr(well[0]) == well[1]
            for well in zip(w6, [well_a1_name, well_b2_name, well_c3_name])
        ]
    )

    w7 = lw.rows("A")
    assert len(w7) == 1
    assert repr(w7[0][0]) == well_a1_name

    w8 = lw.rows("A", "C")
    assert len(w8) == 2
    assert repr(w8[0][0]) == well_a1_name
    assert repr(w8[1][2]) == well_c3_name

    w11 = lw.columns("2", "3", "6")
    assert len(w11) == 3
    assert repr(w11[1][2]) == well_c3_name


def test_well_parent(corning_96_wellplate_360ul_flat: Labware) -> None:
    lw = corning_96_wellplate_360ul_flat
    parent = Location(Point(7, 8, 9), lw)
    well_name = "circular_well_json"
    has_tip = True
    well = labware.Well(
        parent=lw,
        core=LegacyWellCore(
            well_geometry=WellGeometry(
                well_props=test_data[well_name],
                parent_point=parent.point,
                parent_object=parent.labware.as_labware()._core,  # type: ignore[arg-type]
            ),
            display_name=well_name,
            has_tip=has_tip,
            name="A1",
        ),
        api_version=APIVersion(2, 13),
    )
    assert well.parent == lw
    assert well.top().labware.object == well
    assert well.top().labware.parent.object == lw
    assert well.bottom().labware.object == well
    assert well.bottom().labware.parent.object == lw
    assert well.center().labware.object == well
    assert well.center().labware.parent.object == lw


def test_tip_tracking_init(
    corning_96_wellplate_360ul_flat: Labware, opentrons_96_tiprack_300ul: Labware
) -> None:
    tiprack = opentrons_96_tiprack_300ul
    assert tiprack.is_tiprack
    for well in tiprack.wells():
        assert well.has_tip

    lw = corning_96_wellplate_360ul_flat
    assert not lw.is_tiprack
    for well in lw.wells():
        assert not well.has_tip


def test_use_tips(opentrons_96_tiprack_300ul: Labware) -> None:
    tiprack = opentrons_96_tiprack_300ul
    well_list = tiprack.wells()

    # Test only using one tip
    tiprack.use_tips(well_list[0])
    assert not well_list[0].has_tip
    for well in well_list[1:]:
        assert well.has_tip

    # Test using a whole column
    tiprack.use_tips(well_list[8], num_channels=8)
    for well in well_list[8:16]:
        assert not well.has_tip
    assert well_list[7].has_tip
    assert well_list[16].has_tip

    # Test using a partial column from the top
    tiprack.use_tips(well_list[16], num_channels=4)
    for well in well_list[16:20]:
        assert not well.has_tip
    for well in well_list[20:24]:
        assert well.has_tip

    # Test using a partial column where the number of tips that get picked up
    # is less than the number of channels (e.g.: an 8-channel pipette picking
    # up 4 tips from the bottom half of a column)
    tiprack.use_tips(well_list[28], num_channels=4)
    for well in well_list[24:28]:
        assert well.has_tip
    for well in well_list[28:32]:
        assert not well.has_tip
    for well in well_list[32:]:
        assert well.has_tip


def test_select_next_tip(
    opentrons_96_tiprack_300ul: Labware,
    opentrons_96_tiprack_300ul_def: LabwareDefinition2,
) -> None:
    tiprack = opentrons_96_tiprack_300ul
    well_list = tiprack.wells()

    next_one = tiprack.next_tip()
    assert next_one == well_list[0]
    next_five = tiprack.next_tip(5)
    assert next_five == well_list[0]
    next_eight = tiprack.next_tip(8)
    assert next_eight == well_list[0]
    next_nine = tiprack.next_tip(9)
    assert next_nine is None

    # A1 tip only has been used
    tiprack.use_tips(well_list[0])

    next_one = tiprack.next_tip()
    assert next_one == well_list[1]
    next_five = tiprack.next_tip(5)
    assert next_five == well_list[1]
    next_eight = tiprack.next_tip(8)
    assert next_eight == well_list[8]

    # 2nd column has also been used
    tiprack.use_tips(well_list[8], num_channels=8)

    next_one = tiprack.next_tip()
    assert next_one == well_list[1]
    next_five = tiprack.next_tip(5)
    assert next_five == well_list[1]
    next_eight = tiprack.next_tip(8)
    assert next_eight == well_list[16]

    # Bottom 4 tips of 1rd column are also used
    tiprack.use_tips(well_list[4], num_channels=4)

    next_one = tiprack.next_tip()
    assert next_one == well_list[1]
    next_three = tiprack.next_tip(3)
    assert next_three == well_list[1]
    next_five = tiprack.next_tip(5)
    assert next_five == well_list[16]
    next_eight = tiprack.next_tip(8)
    assert next_eight == well_list[16]

    # you can reuse tips infinitely on api level 2.2
    tiprack.use_tips(well_list[0])
    tiprack.use_tips(well_list[0])


def test_previous_tip(opentrons_96_tiprack_300ul: Labware) -> None:
    tiprack = opentrons_96_tiprack_300ul
    # If all wells are used, we can't get a previous tip
    assert tiprack.previous_tip() is None
    # If one well is empty, wherever it is, we can get a slot
    tiprack.wells()[5].has_tip = False
    assert tiprack.previous_tip() == tiprack.wells()[5]
    # But not if we ask for more slots than are available
    assert tiprack.previous_tip(2) is None
    tiprack.wells()[7].has_tip = False
    # And those available wells have to be contiguous
    assert tiprack.previous_tip(2) is None
    # But if they are, we're good
    tiprack.wells()[6].has_tip = False
    assert tiprack.previous_tip(3) == tiprack.wells()[5]


def test_return_tips(opentrons_96_tiprack_300ul: Labware) -> None:
    tiprack = opentrons_96_tiprack_300ul

    # If all wells are used, we get an error if we try to return
    with pytest.raises(AssertionError):
        tiprack.return_tips(tiprack.wells()[0])
    # If we have space where we specify, everything is OK
    tiprack.wells()[0].has_tip = False
    tiprack.return_tips(tiprack.wells()[0])
    assert tiprack.wells()[0].has_tip
    # We have to have enough space
    tiprack.wells()[0].has_tip = False
    with pytest.raises(AssertionError):
        tiprack.return_tips(tiprack.wells()[0], 2)
    # But we can drop stuff off the end of a column
    tiprack.wells()[7].has_tip = False
    tiprack.wells()[8].has_tip = False
    tiprack.return_tips(tiprack.wells()[7], 2)
    assert tiprack.wells()[7].has_tip
    # But we won't wrap around
    assert not tiprack.wells()[8].has_tip


@pytest.mark.parametrize(
    "module_model",
    (
        list(MagneticModuleModel)
        + list(TemperatureModuleModel)
        + list(ThermocyclerModuleModel)
        + list(HeaterShakerModuleModel)
    ),
)
def test_module_geometry_load(module_model: ModuleModel) -> None:
    definition = module_geometry.load_definition(module_model)
    geometry = module_geometry.create_geometry(
        definition=definition,
        parent=Location(Point(0, 0, 0), "3"),
        configuration=None,
    )
    high_z = definition["dimensions"]["bareOverallHeight"]

    assert geometry.highest_z == high_z


@pytest.mark.parametrize(
    "module_name",
    [
        "tempdeck",
        "magdeck",
        "temperature module",
        "temperature module gen2",
        "thermocycler",
        "thermocycler module",
        "magnetic module",
        "magnetic module gen2",
        "magneticModuleV1",
        "magneticModuleV2",
        "temperatureModuleV1",
        "temperatureModuleV2",
        "thermocyclerModuleV1",
        "thermocyclerModuleV2",
    ],
)
def test_module_load_labware(module_name: str) -> None:
    labware_name = "corning_96_wellplate_360ul_flat"
    labware_def = labware.get_labware_definition(labware_name)
    mod_model = validation.ensure_module_model(module_name)
    mod_definition = module_geometry.load_definition(mod_model)
    mod = module_geometry.create_geometry(
        mod_definition,
        Location(Point(0, 0, 0), "test"),
        None,
    )
    old_z = mod.highest_z
    assert labware_def["schemaVersion"] == 2  # load_from_definition() expects this.
    lw = labware.load_from_definition(labware_def, mod.location)
    mod.add_labware(lw)
    assert mod.labware == lw
    assert mod.highest_z == (
        mod.location.point.z
        + labware_def["dimensions"]["zDimension"]
        + mod._over_labware
    )
    with pytest.raises(RuntimeError):
        mod.add_labware(lw)
    mod.reset_labware()
    assert mod.labware is None
    assert mod.highest_z == old_z


def test_tiprack_list() -> None:
    labware_name = "opentrons_96_tiprack_300ul"
    labware_def = labware.get_labware_definition(labware_name)
    assert labware_def["schemaVersion"] == 2  # LegacyLabwareCore expects this.
    tiprack = labware.Labware(
        core=LegacyLabwareCore(labware_def, Location(Point(0, 0, 0), "Test Slot")),
        api_version=APIVersion(2, 13),
        protocol_core=None,  # type: ignore[arg-type]
        core_map=None,  # type: ignore[arg-type]
    )
    tiprack_2 = labware.Labware(
        core=LegacyLabwareCore(labware_def, Location(Point(0, 0, 0), "Test Slot")),
        api_version=APIVersion(2, 13),
        protocol_core=None,  # type: ignore[arg-type]
        core_map=None,  # type: ignore[arg-type]
    )

    assert labware.select_tiprack_from_list([tiprack], 1) == (
        tiprack,
        tiprack["A1"],
    )

    assert labware.select_tiprack_from_list([tiprack], 1, tiprack.wells()[1]) == (
        tiprack,
        tiprack["B1"],
    )

    tiprack["C1"].has_tip = False
    assert labware.select_tiprack_from_list([tiprack], 1, tiprack.wells()[2]) == (
        tiprack,
        tiprack["D1"],
    )

    tiprack["H12"].has_tip = False
    tiprack_2["A1"].has_tip = False
    assert labware.select_tiprack_from_list(
        [tiprack, tiprack_2], 1, tiprack.wells()[95]
    ) == (tiprack_2, tiprack_2["B1"])

    with pytest.raises(labware.OutOfTipsError):
        labware.select_tiprack_from_list([tiprack], 1, tiprack.wells()[95])


def test_uris() -> None:
    details = ("opentrons", "opentrons_96_tiprack_300ul", "1")
    uri = "opentrons/opentrons_96_tiprack_300ul/1"
    assert helpers.uri_from_details(*details) == uri
    defn = labware.get_labware_definition(
        details[1],
        details[0],
        details[2],  # type: ignore[arg-type]
    )
    assert defn["schemaVersion"] == 2  # LegacyLabwareCore expects this.
    assert helpers.uri_from_definition(defn) == uri
    lw = labware.Labware(
        core=LegacyLabwareCore(defn, Location(Point(0, 0, 0), "Test Slot")),
        api_version=APIVersion(2, 13),
        protocol_core=None,  # type: ignore[arg-type]
        core_map=None,  # type: ignore[arg-type]
    )
    assert lw.uri == uri


def test_labware_hash_func_same_implementation(
    minimal_labware_def: LabwareDefinition2,
) -> None:
    """Test that multiple Labware objects with same implementation and version
    have the same __hash__"""
    impl = LegacyLabwareCore(minimal_labware_def, Location(Point(0, 0, 0), "Test Slot"))
    s = set(
        labware.Labware(
            core=impl,
            api_version=APIVersion(2, 3),
            protocol_core=None,  # type: ignore[arg-type]
            core_map=None,  # type: ignore[arg-type]
        )
        for i in range(10)
    )
    assert len(s) == 1


def test_labware_hash_func_same_implementation_different_version(
    minimal_labware_def: LabwareDefinition2,
) -> None:
    """Test that multiple Labware objects with same implementation yet
    different version have different __hash__"""
    impl = LegacyLabwareCore(minimal_labware_def, Location(Point(0, 0, 0), "Test Slot"))

    l1 = labware.Labware(
        core=impl,
        api_version=APIVersion(2, 13),
        protocol_core=None,  # type: ignore[arg-type]
        core_map=None,  # type: ignore[arg-type]
    )
    l2 = labware.Labware(
        core=impl,
        api_version=APIVersion(2, 14),
        protocol_core=None,  # type: ignore[arg-type]
        core_map=None,  # type: ignore[arg-type]
    )

    assert len({l1, l2}) == 2


def test_labware_hash_func_diff_implementation_same_version(
    minimal_labware_def: LabwareDefinition2,
) -> None:
    """Test that multiple Labware objects with different implementation yet
    sane version have different __hash__"""
    impl1 = LegacyLabwareCore(
        minimal_labware_def, Location(Point(0, 0, 0), "Test Slot")
    )
    impl2 = LegacyLabwareCore(
        minimal_labware_def, Location(Point(0, 0, 0), "Test Slot2")
    )

    l1 = labware.Labware(
        core=impl1,
        api_version=APIVersion(2, 3),
        protocol_core=None,  # type: ignore[arg-type]
        core_map=None,  # type: ignore[arg-type]
    )
    l2 = labware.Labware(
        core=impl2,
        api_version=APIVersion(2, 3),
        protocol_core=None,  # type: ignore[arg-type]
        core_map=None,  # type: ignore[arg-type]
    )

    assert len({l1, l2}) == 2


def test_set_offset(decoy: Decoy) -> None:
    """It should set the labware's offset using the implementation."""
    labware_impl: AbstractLabware[AbstractWellCore] = decoy.mock(cls=AbstractLabware)
    decoy.when(labware_impl.get_well_columns()).then_return([])
    subject = labware.Labware(
        core=labware_impl,
        api_version=APIVersion(2, 12),
        protocol_core=None,  # type: ignore[arg-type]
        core_map=None,  # type: ignore[arg-type]
    )

    subject.set_offset(x=1.1, y=2.2, z=3.3)
    decoy.verify(labware_impl.set_calibration(Point(1.1, 2.2, 3.3)))



"""Vacuum Module-specific errors and exceptions."""

from opentrons.drivers.asyncio.communication.errors import (
    BaseErrorCode,
    ErrorResponse,
    GCodeCacheFull,
    TaskNotReady,
    UnhandledGcode,
)


class EStopTriggered(ErrorResponse):
    """Raised when the estop is triggered during a module action."""

    def __init__(self, port: str, response: str, command: str) -> None:
        super().__init__(port, response, command)


class PumpMotorError(ErrorResponse):
    """Raised when pump motor error is received."""

    def __init__(self, port: str, response: str, command: str) -> None:
        super().__init__(port, response, command)


class StopRequested(ErrorResponse):
    """Raised when a stop is requested during a module action."""

    def __init__(self, port: str, response: str, command: str) -> None:
        super().__init__(port, response, command)


class VacuumModuleErrorCodes(BaseErrorCode):
    """Vacuum Module Error Codes."""

    UNHANDLED_GCODE = ("ERR003", UnhandledGcode)
    GCODE_CACHE_FULL = ("ERR004", GCodeCacheFull)
    TASK_NOT_READY = ("ERR007", TaskNotReady)
    STOP_REQUESTED = ("ERR504", StopRequested)



import logging
from typing import Optional

import pytest

from opentrons.protocol_api import ProtocolContext
from opentrons.protocols.api_support.instrument import (
    validate_takes_liquid,
    validate_tiprack,
)
from opentrons.types import Location, Point


@pytest.mark.parametrize("reject_module", [True, False])
def test_validate_takes_liquid(ctx: ProtocolContext, reject_module: bool) -> None:
    well_plate = ctx.load_labware("corning_96_wellplate_360ul_flat", 1)
    tip_rack = ctx.load_labware("opentrons_96_tiprack_300ul", 2)

    validate_takes_liquid(
        location=Location(Point(1, 2, 3), None),
        reject_module=reject_module,
        reject_adapter=False,
    )
    validate_takes_liquid(
        location=Location(Point(1, 2, 3), well_plate),
        reject_module=reject_module,
        reject_adapter=False,
    )
    validate_takes_liquid(
        location=Location(Point(1, 2, 3), well_plate.wells()[0]),
        reject_module=reject_module,
        reject_adapter=False,
    )
    validate_takes_liquid(
        location=well_plate.wells()[0].top(),
        reject_module=reject_module,
        reject_adapter=False,
    )

    with pytest.raises(ValueError, match="Cannot aspirate/dispense to a tip rack"):
        validate_takes_liquid(
            location=Location(Point(1, 2, 3), tip_rack),
            reject_module=reject_module,
            reject_adapter=False,
        )

    with pytest.raises(ValueError, match="Cannot aspirate/dispense to a tip rack"):
        validate_takes_liquid(
            location=Location(Point(1, 2, 3), tip_rack.wells()[0]),
            reject_module=reject_module,
            reject_adapter=False,
        )

    with pytest.raises(ValueError, match="Cannot aspirate/dispense to a tip rack"):
        validate_takes_liquid(
            location=tip_rack.wells_by_name()["A1"].top(),
            reject_module=reject_module,
            reject_adapter=False,
        )


# TODO(mm, 2023-04-28): The validate_takes_liquid() function is used both by ProtocolContexts
# that are backed by Protocol Engine, and those that aren't. But this test is only runnable
# with a non-Protocol-Engine ProtocolContext because it relies on the internal module.geometry
# property.
#
# Find a different way to test this so that both paths are covered.
@pytest.mark.apiv2_non_pe_only
def test_validate_takes_liquid_module_location(ctx: ProtocolContext) -> None:
    module = ctx.load_module("magdeck", 1)

    validate_takes_liquid(
        location=module.geometry.location,
        reject_module=False,
        reject_adapter=False,
    )

    with pytest.raises(
        ValueError,
        match="Cannot aspirate/dispense directly to a module",
    ):
        validate_takes_liquid(
            location=module.geometry.location,
            reject_module=True,
            reject_adapter=False,
        )


@pytest.mark.ot3_only
def test_validate_takes_liquid_adapter(ctx: ProtocolContext) -> None:
    well_plate = ctx.load_labware("corning_96_wellplate_360ul_flat", 1)
    adapter = ctx.load_adapter("opentrons_96_pcr_adapter", 2)

    validate_takes_liquid(
        location=Location(Point(1, 2, 3), None),
        reject_module=False,
        reject_adapter=True,
    )
    validate_takes_liquid(
        location=Location(Point(1, 2, 3), well_plate),
        reject_module=False,
        reject_adapter=True,
    )
    validate_takes_liquid(
        location=Location(Point(1, 2, 3), well_plate.wells()[0]),
        reject_module=False,
        reject_adapter=True,
    )
    validate_takes_liquid(
        location=well_plate.wells()[0].top(),
        reject_module=False,
        reject_adapter=True,
    )

    with pytest.raises(ValueError, match="Cannot aspirate/dispense to an adapter"):
        validate_takes_liquid(
            location=Location(Point(1, 2, 3), adapter),
            reject_module=False,
            reject_adapter=True,
        )

    with pytest.raises(ValueError, match="Cannot aspirate/dispense to an adapter"):
        validate_takes_liquid(
            location=Location(Point(1, 2, 3), adapter.wells()[0]),
            reject_module=False,
            reject_adapter=True,
        )

    with pytest.raises(ValueError, match="Cannot aspirate/dispense to an adapter"):
        validate_takes_liquid(
            location=adapter.wells_by_name()["A1"].top(),
            reject_module=False,
            reject_adapter=True,
        )


@pytest.mark.parametrize(
    argnames=["pipette_name", "log_value"],
    argvalues=[
        ["p1000_96", None],
        [
            "p50_single_flex",
            "The pipette p50_single_flex and its tip rack opentrons_flex_96_tiprack_200ul appear to be mismatched. Please check your protocol.",
        ],
        [
            "p20_single_gen2",
            "The pipette p20_single_gen2 and its tip rack opentrons_flex_96_tiprack_200ul appear to be mismatched. Please check your protocol.",
        ],
    ],
)
def test_validate_tiprack(
    ctx: ProtocolContext,
    caplog: pytest.LogCaptureFixture,
    pipette_name: str,
    log_value: Optional[str],
) -> None:
    tip_rack = ctx.load_labware("opentrons_flex_96_tiprack_200ul", 2)

    with caplog.at_level(logging.WARNING):
        log = logging.getLogger(__name__)
        validate_tiprack(pipette_name, tip_rack, log=log)

    if log_value:
        assert caplog.messages[0] == log_value
    else:
        assert caplog.records == []



"""Test the functionality of the `RobotContext`."""

from typing import Optional, Union

import pytest
from decoy import Decoy

from opentrons_shared_data.pipette.types import PipetteNameType

from opentrons.protocol_api import ModuleContext, RobotContext
from opentrons.protocol_api._types import PipetteActionTypes, PlungerPositionTypes
from opentrons.protocol_api.core.common import ProtocolCore, RobotCore
from opentrons.protocol_api.deck import Deck
from opentrons.protocols.api_support.types import APIVersion
from opentrons.types import (
    AxisMapType,
    AxisType,
    DeckLocation,
    DeckSlotName,
    Location,
    Mount,
    Point,
    StringAxisMap,
)


@pytest.fixture
def mock_core(decoy: Decoy) -> RobotCore:
    """Get a mock module implementation core."""
    return decoy.mock(cls=RobotCore)


@pytest.fixture
def api_version() -> APIVersion:
    """Get the API version to test at."""
    return APIVersion(2, 22)


@pytest.fixture
def mock_deck(decoy: Decoy) -> Deck:
    """Get a mocked deck object."""
    deck = decoy.mock(cls=Deck)
    decoy.when(deck.get_slot_center(DeckSlotName.SLOT_D1.value)).then_return(
        Point(3, 3, 3)
    )
    return deck


@pytest.fixture
def mock_protocol(decoy: Decoy, mock_deck: Deck, mock_core: RobotCore) -> ProtocolCore:
    """Get a mock protocol implementation core without a 96 channel attached."""
    protocol_core = decoy.mock(cls=ProtocolCore)
    decoy.when(protocol_core.robot_type).then_return("OT-3 Standard")
    decoy.when(protocol_core.load_robot()).then_return(mock_core)
    return protocol_core


@pytest.fixture
def subject(
    decoy: Decoy,
    mock_core: RobotCore,
    mock_protocol: ProtocolCore,
    api_version: APIVersion,
) -> RobotContext:
    """Get a RobotContext test subject with its dependencies mocked out."""
    decoy.when(mock_core.get_pipette_type_from_engine(Mount.LEFT)).then_return(
        PipetteNameType.P1000_SINGLE_FLEX
    )
    decoy.when(mock_core.get_pipette_type_from_engine(Mount.RIGHT)).then_return(
        PipetteNameType.P1000_SINGLE_FLEX
    )
    return RobotContext(
        core=mock_core, api_version=api_version, protocol_core=mock_protocol
    )


@pytest.mark.parametrize(
    argnames=["mount", "destination", "speed"],
    argvalues=[
        ("left", Location(point=Point(1, 2, 3), labware=None), None),
        (Mount.RIGHT, Location(point=Point(1, 2, 3), labware=None), 100),
    ],
)
def test_move_to(
    decoy: Decoy,
    subject: RobotContext,
    mount: Union[str, Mount],
    destination: Location,
    speed: Optional[float],
) -> None:
    """Test `RobotContext.move_to`."""
    subject.move_to(mount, destination, speed)
    core_mount: Mount
    if isinstance(mount, str):
        core_mount = Mount.string_to_mount(mount)
    else:
        core_mount = mount
    decoy.verify(subject._core.move_to(core_mount, destination.point, speed))


@pytest.mark.parametrize(
    argnames=[
        "axis_map",
        "critical_point",
        "expected_axis_map",
        "expected_critical_point",
        "speed",
    ],
    argvalues=[
        (
            {"x": 100, "Y": 50, "z_g": 80},
            {"x": 5, "Y": 5, "z_g": 5},
            {AxisType.X: 100, AxisType.Y: 50, AxisType.Z_G: 80},
            {AxisType.X: 5, AxisType.Y: 5, AxisType.Z_G: 5},
            None,
        ),
        (
            {"x": 5, "Y": 5},
            {"x": 5, "Y": 5},
            {AxisType.X: 5, AxisType.Y: 5},
            {AxisType.X: 5, AxisType.Y: 5},
            None,
        ),
    ],
)
def test_move_axes_to(
    decoy: Decoy,
    subject: RobotContext,
    axis_map: Union[StringAxisMap, AxisMapType],
    critical_point: Union[StringAxisMap, AxisMapType],
    expected_axis_map: AxisMapType,
    expected_critical_point: AxisMapType,
    speed: Optional[float],
) -> None:
    """Test `RobotContext.move_axes_to`."""
    subject.move_axes_to(axis_map, critical_point, speed)
    decoy.verify(
        subject._core.move_axes_to(expected_axis_map, expected_critical_point, speed)
    )


@pytest.mark.parametrize(
    argnames=["axis_map", "converted_map", "speed"],
    argvalues=[
        (
            {"x": 10, "Y": 10, "z_g": 10},
            {AxisType.X: 10, AxisType.Y: 10, AxisType.Z_G: 10},
            None,
        ),
        ({AxisType.P_L: 10}, {AxisType.P_L: 10}, 5),
    ],
)
def test_move_axes_relative(
    decoy: Decoy,
    subject: RobotContext,
    axis_map: Union[StringAxisMap, AxisMapType],
    converted_map: AxisMapType,
    speed: Optional[float],
) -> None:
    """Test `RobotContext.move_axes_relative`."""
    subject.move_axes_relative(axis_map, speed)
    decoy.verify(subject._core.move_axes_relative(converted_map, speed))


@pytest.mark.parametrize(
    argnames=["mount", "location_to_move", "expected_axis_map"],
    argvalues=[
        (
            "left",
            Location(point=Point(1, 2, 3), labware=None),
            {AxisType.Z_L: 3, AxisType.X: 1, AxisType.Y: 2},
        ),
        (
            Mount.EXTENSION,
            Location(point=Point(1, 2, 3), labware=None),
            {AxisType.Z_G: 3, AxisType.X: 1, AxisType.Y: 2},
        ),
    ],
)
def test_get_axes_coordinates_for(
    subject: RobotContext,
    mount: Union[Mount, str],
    location_to_move: Union[Location, ModuleContext, DeckLocation],
    expected_axis_map: AxisMapType,
) -> None:
    """Test `RobotContext.get_axis_coordinates_for`."""
    res = subject.axis_coordinates_for(mount, location_to_move)
    assert res == expected_axis_map


@pytest.mark.parametrize(
    argnames=["mount", "volume", "action", "expected_axis_map"],
    argvalues=[
        (Mount.RIGHT, 200, PipetteActionTypes.ASPIRATE_ACTION, {AxisType.P_R: 100}),
        (Mount.LEFT, 100, PipetteActionTypes.DISPENSE_ACTION, {AxisType.P_L: 100}),
    ],
)
def test_plunger_coordinates_for_volume(
    decoy: Decoy,
    subject: RobotContext,
    mount: Mount,
    volume: float,
    action: PipetteActionTypes,
    expected_axis_map: AxisMapType,
) -> None:
    """Test `RobotContext.plunger_coordinates_for_volume`."""
    decoy.when(
        subject._core.get_plunger_position_from_volume(
            mount, volume, action, "OT-3 Standard"
        )
    ).then_return(100)

    result = subject.plunger_coordinates_for_volume(mount, volume, action)
    assert result == expected_axis_map


@pytest.mark.parametrize(
    argnames=["mount", "position_name", "expected_axis_map"],
    argvalues=[
        (Mount.RIGHT, PlungerPositionTypes.PLUNGER_TOP, {AxisType.P_R: 3}),
        (
            Mount.RIGHT,
            PlungerPositionTypes.PLUNGER_BOTTOM,
            {AxisType.P_R: 3},
        ),
    ],
)
def test_plunger_coordinates_for_named_position(
    decoy: Decoy,
    subject: RobotContext,
    mount: Mount,
    position_name: PlungerPositionTypes,
    expected_axis_map: AxisMapType,
) -> None:
    """Test `RobotContext.plunger_coordinates_for_named_position`."""
    decoy.when(
        subject._core.get_plunger_position_from_name(mount, position_name)
    ).then_return(3)
    result = subject.plunger_coordinates_for_named_position(mount, position_name)
    assert result == expected_axis_map


def test_plunger_methods_raise_without_pipette(
    mock_core: RobotCore, mock_protocol: ProtocolCore, api_version: APIVersion
) -> None:
    """Test that `RobotContext` plunger functions raise without pipette attached."""
    subject = RobotContext(
        core=mock_core, api_version=api_version, protocol_core=mock_protocol
    )
    with pytest.raises(ValueError):
        subject.plunger_coordinates_for_named_position(
            Mount.LEFT, PlungerPositionTypes.PLUNGER_TOP
        )

    with pytest.raises(ValueError):
        subject.plunger_coordinates_for_volume(
            Mount.LEFT, 200, PipetteActionTypes.ASPIRATE_ACTION
        )



"""Tests for Protocol API Flex Stacker contexts."""

from unittest.mock import sentinel

import pytest
from decoy import Decoy, matchers

from . import versions_at_or_above, versions_between
from opentrons.legacy_broker import LegacyBroker
from opentrons.protocol_api import FlexStackerContext, Labware
from opentrons.protocol_api.core.common import (
    FlexStackerCore,
    LabwareCore,
    ProtocolCore,
)
from opentrons.protocol_api.core.core_map import LoadedCoreMap
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.api_support.util import APIVersionError
from opentrons.types import DeckSlotName


@pytest.fixture
def mock_core(decoy: Decoy) -> FlexStackerCore:
    """Get a mock module implementation core."""
    core = decoy.mock(cls=FlexStackerCore)
    decoy.when(core.get_display_name()).then_return("mock stacker core")
    decoy.when(core.get_deck_slot()).then_return(DeckSlotName.SLOT_D3)
    return core


@pytest.fixture
def mock_protocol_core(decoy: Decoy) -> ProtocolCore:
    """Get a mock protocol implementation core."""
    return decoy.mock(cls=ProtocolCore)


@pytest.fixture
def mock_labware_core(decoy: Decoy) -> LabwareCore:
    """Get a mock labware implementation core."""
    mock_core = decoy.mock(cls=LabwareCore)
    decoy.when(mock_core.get_well_columns()).then_return([])
    return mock_core


@pytest.fixture
def mock_core_map(decoy: Decoy) -> LoadedCoreMap:
    """Get a mock LoadedCoreMap."""
    return decoy.mock(cls=LoadedCoreMap)


@pytest.fixture
def mock_broker(decoy: Decoy) -> LegacyBroker:
    """Get a mock command message broker."""
    return decoy.mock(cls=LegacyBroker)


@pytest.fixture
def subject(
    api_version: APIVersion,
    mock_core: FlexStackerCore,
    mock_protocol_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    mock_broker: LegacyBroker,
) -> FlexStackerContext:
    """Get an absorbance reader context with its dependencies mocked out."""
    return FlexStackerContext(
        core=mock_core,
        protocol_core=mock_protocol_core,
        core_map=mock_core_map,
        broker=mock_broker,
        api_version=api_version,
    )


@pytest.mark.parametrize(
    "api_version", versions_at_or_above(from_version=APIVersion(2, 25))
)
def test_get_serial_number(
    decoy: Decoy, mock_core: FlexStackerCore, subject: FlexStackerContext
) -> None:
    """It should get the serial number from the core."""
    decoy.when(mock_core.get_serial_number()).then_return("12345")
    result = subject.serial_number
    assert result == "12345"


@pytest.mark.parametrize(
    "api_version", versions_at_or_above(from_version=APIVersion(2, 25))
)
def test_fill(
    decoy: Decoy, mock_core: FlexStackerCore, subject: FlexStackerContext
) -> None:
    """It should pass args to the core."""
    subject.fill(2, "hello")
    decoy.verify(mock_core.fill(2, "hello"))


@pytest.mark.parametrize(
    "api_version", versions_at_or_above(from_version=APIVersion(2, 25))
)
def test_empty(
    decoy: Decoy, mock_core: FlexStackerCore, subject: FlexStackerContext
) -> None:
    """It should pass args to the core."""
    subject.empty("goodbye")
    decoy.verify(mock_core.empty("goodbye"))


@pytest.mark.parametrize(
    "api_version", versions_at_or_above(from_version=APIVersion(2, 26))
)
def test_set_stored_labware(
    decoy: Decoy, mock_core: FlexStackerCore, subject: FlexStackerContext
) -> None:
    """It should route arguments appropriately."""
    subject.set_stored_labware(
        load_name=sentinel.main_load_name,
        namespace=sentinel.main_namespace,
        version=sentinel.main_version,
        adapter=sentinel.adapter_load_name,
        lid=sentinel.lid_load_name,
        count=sentinel.count,
        stacking_offset_z=sentinel.stacking_offset_z,
    )
    decoy.verify(
        mock_core.set_stored_labware(
            main_load_name=sentinel.main_load_name,
            main_namespace=sentinel.main_namespace,
            main_version=sentinel.main_version,
            lid_load_name=sentinel.lid_load_name,
            lid_namespace=None,
            lid_version=None,
            adapter_load_name=sentinel.adapter_load_name,
            adapter_namespace=None,
            adapter_version=None,
            count=sentinel.count,
            stacking_offset_z=sentinel.stacking_offset_z,
        )
    )

    subject.set_stored_labware(
        load_name=sentinel.main_load_name,
        namespace=sentinel.main_namespace,
        version=sentinel.main_version,
        adapter=sentinel.adapter_load_name,
        lid=sentinel.lid_load_name,
        count=sentinel.count,
        stacking_offset_z=sentinel.stacking_offset_z,
        adapter_namespace=sentinel.adapter_namespace,
        adapter_version=sentinel.adapter_version,
        lid_namespace=sentinel.lid_namespace,
        lid_version=sentinel.lid_version,
    )
    decoy.verify(
        mock_core.set_stored_labware(
            main_load_name=sentinel.main_load_name,
            main_namespace=sentinel.main_namespace,
            main_version=sentinel.main_version,
            lid_load_name=sentinel.lid_load_name,
            lid_namespace=sentinel.lid_namespace,
            lid_version=sentinel.lid_version,
            adapter_load_name=sentinel.adapter_load_name,
            adapter_namespace=sentinel.adapter_namespace,
            adapter_version=sentinel.adapter_version,
            count=sentinel.count,
            stacking_offset_z=sentinel.stacking_offset_z,
        )
    )


@pytest.mark.parametrize(
    "api_version",
    versions_between(
        low_inclusive_bound=APIVersion(2, 25), high_exclusive_bound=APIVersion(2, 26)
    ),
)
def test_set_stored_labware_namespace_version_params_minimum_version(
    subject: FlexStackerContext,
) -> None:
    """{adapter,lid}_{namespace,version} params were unavailable before version 2.26."""
    with pytest.raises(APIVersionError):
        subject.set_stored_labware(
            "load_name",
            "namespace",
            1,
            "adapter",
            "lid",
            2,
            stacking_offset_z=1.0,
            adapter_namespace="foo",
        )
    with pytest.raises(APIVersionError):
        subject.set_stored_labware(
            "load_name",
            "namespace",
            1,
            "adapter",
            "lid",
            2,
            stacking_offset_z=1.0,
            adapter_version=123,
        )
    with pytest.raises(APIVersionError):
        subject.set_stored_labware(
            "load_name",
            "namespace",
            1,
            "adapter",
            "lid",
            2,
            stacking_offset_z=1.0,
            lid_namespace="foo",
        )
    with pytest.raises(APIVersionError):
        subject.set_stored_labware(
            "load_name",
            "namespace",
            1,
            "adapter",
            "lid",
            2,
            stacking_offset_z=1.0,
            lid_version=123,
        )


@pytest.mark.parametrize(
    "api_version",
    versions_between(
        low_inclusive_bound=APIVersion(2, 25), high_exclusive_bound=APIVersion(2, 26)
    ),
)
def test_set_stored_labware_namespace_version_params_legacy_behavior(
    decoy: Decoy,
    mock_core: FlexStackerCore,
    subject: FlexStackerContext,
) -> None:
    """Test legacy (version <2.26) behavior when the {adapter,lid}_{namespace,version} params were omitted.

    The adapter and lid namespace and version were taken from the main labware's namespace and version.
    """
    subject.set_stored_labware(
        load_name=sentinel.main_load_name,
        namespace=sentinel.main_namespace,
        version=sentinel.main_version,
        adapter=sentinel.adapter_load_name,
        lid=sentinel.lid_load_name,
        count=sentinel.count,
        stacking_offset_z=sentinel.stacking_offset_z,
    )

    decoy.verify(
        mock_core.set_stored_labware(
            main_load_name=sentinel.main_load_name,
            main_namespace=sentinel.main_namespace,
            main_version=sentinel.main_version,
            lid_load_name=sentinel.lid_load_name,
            lid_namespace=sentinel.main_namespace,
            lid_version=sentinel.main_version,
            adapter_load_name=sentinel.adapter_load_name,
            adapter_namespace=sentinel.main_namespace,
            adapter_version=sentinel.main_version,
            count=sentinel.count,
            stacking_offset_z=sentinel.stacking_offset_z,
        )
    )


@pytest.mark.parametrize(
    "api_version", versions_at_or_above(from_version=APIVersion(2, 25))
)
def test_get_max_storable_labware_from_list(
    decoy: Decoy,
    mock_core: FlexStackerCore,
    mock_protocol_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    subject: FlexStackerContext,
) -> None:
    """It should filter its arguments and responses."""
    base_cores = [decoy.mock(cls=LabwareCore) for _ in range(5)]
    for idx, core in enumerate(base_cores):
        decoy.when(core.get_well_columns()).then_return([])
        decoy.when(core.get_display_name()).then_return(f"core-{idx}")
        decoy.when(mock_core_map.get_or_add(core, matchers.Anything())).then_do(
            lambda lw, builder: builder(lw)
        )
    base_lw = [
        Labware(
            core=core,
            api_version=APIVersion(2, 25),
            protocol_core=mock_protocol_core,
            core_map=subject._core_map,
        )
        for core in base_cores
    ]
    decoy.when(
        mock_core.get_max_storable_labware_from_list(base_cores, 1.0)
    ).then_return(base_cores[:3])
    assert subject.get_max_storable_labware_from_list(base_lw, 1.0) == base_lw[:3]


@pytest.mark.parametrize(
    "api_version", versions_at_or_above(from_version=APIVersion(2, 25))
)
def test_get_current_storable_labware_from_list(
    decoy: Decoy,
    mock_core: FlexStackerCore,
    mock_protocol_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    subject: FlexStackerContext,
) -> None:
    """It should filter its arguments and responses."""
    base_cores = [decoy.mock(cls=LabwareCore) for _ in range(5)]
    for idx, core in enumerate(base_cores):
        decoy.when(core.get_well_columns()).then_return([])
        decoy.when(core.get_display_name()).then_return(f"core-{idx}")
        decoy.when(mock_core_map.get_or_add(core, matchers.Anything())).then_do(
            lambda lw, builder: builder(lw)
        )
    base_lw = [
        Labware(
            core=core,
            api_version=APIVersion(2, 25),
            protocol_core=mock_protocol_core,
            core_map=subject._core_map,
        )
        for core in base_cores
    ]
    decoy.when(
        mock_core.get_current_storable_labware_from_list(base_cores)
    ).then_return(base_cores[:3])
    assert subject.get_current_storable_labware_from_list(base_lw) == base_lw[:3]


@pytest.mark.parametrize(
    "api_version", versions_at_or_above(from_version=APIVersion(2, 25))
)
def test_get_max_storable_labware(
    decoy: Decoy,
    mock_core: FlexStackerCore,
    subject: FlexStackerContext,
) -> None:
    """It should filter its arguments and responses."""
    decoy.when(mock_core.get_max_storable_labware()).then_return(3)
    assert subject.get_max_storable_labware() == 3


@pytest.mark.parametrize(
    "api_version", versions_at_or_above(from_version=APIVersion(2, 25))
)
def test_get_current_storable_labware(
    decoy: Decoy,
    mock_core: FlexStackerCore,
    subject: FlexStackerContext,
) -> None:
    """It should filter its arguments and responses."""
    decoy.when(mock_core.get_current_storable_labware()).then_return(3)
    assert subject.get_current_storable_labware() == 3


@pytest.mark.parametrize(
    "api_version", versions_at_or_above(from_version=APIVersion(2, 25))
)
def test_get_stored_labware(
    decoy: Decoy,
    mock_core: FlexStackerCore,
    mock_protocol_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    subject: FlexStackerContext,
) -> None:
    """It should wrap the response in Labwares."""
    base_cores = [decoy.mock(cls=LabwareCore) for _ in range(5)]
    for idx, core in enumerate(base_cores):
        decoy.when(core.get_well_columns()).then_return([])
        decoy.when(core.get_display_name()).then_return(f"core-{idx}")
        decoy.when(mock_core_map.get_or_add(core, matchers.Anything())).then_do(
            lambda lw, builder: builder(lw)
        )
    base_lw = [
        Labware(
            core=core,
            api_version=APIVersion(2, 25),
            protocol_core=mock_protocol_core,
            core_map=subject._core_map,
        )
        for core in base_cores
    ]
    decoy.when(mock_core.get_stored_labware()).then_return(base_cores)
    assert subject.get_stored_labware() == base_lw


@pytest.mark.parametrize(
    "api_version", versions_at_or_above(from_version=APIVersion(2, 25))
)
@pytest.mark.parametrize("message", ["hello", None])
def test_fill_items(
    decoy: Decoy,
    mock_core: FlexStackerCore,
    mock_protocol_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    subject: FlexStackerContext,
    message: str | None,
) -> None:
    """It should wrap the response in Labwares."""
    base_cores = [decoy.mock(cls=LabwareCore) for _ in range(5)]
    for idx, core in enumerate(base_cores):
        decoy.when(core.get_well_columns()).then_return([])
        decoy.when(core.get_display_name()).then_return(f"core-{idx}")

    base_lw = [
        Labware(
            core=core,
            api_version=APIVersion(2, 25),
            protocol_core=mock_protocol_core,
            core_map=subject._core_map,
        )
        for core in base_cores
    ]
    subject.fill_items(base_lw, message)
    decoy.verify(mock_core.fill_items(base_cores, message))


@pytest.mark.parametrize(
    "api_version", versions_at_or_above(from_version=APIVersion(2, 25))
)
def test_set_stored_labware_items(
    decoy: Decoy,
    mock_core: FlexStackerCore,
    mock_protocol_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    subject: FlexStackerContext,
) -> None:
    """It should wrap the response in Labwares."""
    base_cores = [decoy.mock(cls=LabwareCore) for _ in range(5)]
    for idx, core in enumerate(base_cores):
        decoy.when(core.get_well_columns()).then_return([])
        decoy.when(core.get_display_name()).then_return(f"core-{idx}")

    base_lw = [
        Labware(
            core=core,
            api_version=APIVersion(2, 25),
            protocol_core=mock_protocol_core,
            core_map=subject._core_map,
        )
        for core in base_cores
    ]
    subject.set_stored_labware_items(base_lw, stacking_offset_z=1.0)
    decoy.verify(mock_core.set_stored_labware_items(base_cores, 1.0))



"""Tests for the WellGrid class."""

from opentrons.protocol_api.core import well_grid as subject


def test_empty_well_grid() -> None:
    """It have empty rows and colums."""
    result = subject.create(columns=[])

    assert result.columns_by_name == {}
    assert result.rows_by_name == {}


def test_well_rows() -> None:
    """It should transpose columns to rows."""
    result = subject.create(
        columns=[
            ["A1", "B1"],
            ["A2", "B2"],
        ]
    )

    assert result.columns_by_name == {
        "1": ["A1", "B1"],
        "2": ["A2", "B2"],
    }

    assert result.rows_by_name == {
        "A": ["A1", "A2"],
        "B": ["B1", "B2"],
    }


def test_well_rows_asymmetric() -> None:
    """It should transpose columns to rows when rows are asymmetrical."""
    result = subject.create(
        columns=[
            ["A1", "B1", "C1"],
            ["A2", "B2", "C2"],
            ["A3", "B3"],
            ["A4", "B4"],
        ]
    )

    assert result.columns_by_name == {
        "1": ["A1", "B1", "C1"],
        "2": ["A2", "B2", "C2"],
        "3": ["A3", "B3"],
        "4": ["A4", "B4"],
    }

    assert result.rows_by_name == {
        "A": ["A1", "A2", "A3", "A4"],
        "B": ["B1", "B2", "B3", "B4"],
        "C": ["C1", "C2"],
    }


def test_well_rows_asymmetric_bottom_aligned() -> None:
    """It should transpose columns to rows when asymmetric rows are bottom aligned."""
    result = subject.create(
        columns=[
            ["A1", "B1", "C1"],
            ["A2", "B2", "C2"],
            ["B3", "C3"],
            ["B4", "C4"],
        ]
    )

    assert result.columns_by_name == {
        "1": ["A1", "B1", "C1"],
        "2": ["A2", "B2", "C2"],
        "3": ["B3", "C3"],
        "4": ["B4", "C4"],
    }

    assert result.rows_by_name == {
        "A": ["A1", "A2"],
        "B": ["B1", "B2", "B3", "B4"],
        "C": ["C1", "C2", "C3", "C4"],
    }



from typing import TYPE_CHECKING, Any

import pytest

from opentrons.calibration_storage import (
    types as cs_types,
)
from opentrons.calibration_storage.ot2 import (
    clear_pipette_offset_calibrations,
    delete_pipette_offset_file,
    get_pipette_offset,
    save_pipette_calibration,
)
from opentrons.types import Mount, Point

if TYPE_CHECKING:
    from opentrons_shared_data.deck.types import RobotModel


@pytest.fixture
def starting_calibration_data(
    robot_model: "RobotModel", ot_config_tempdir: Any
) -> None:
    """
    Starting calibration data fixture.

    Adds dummy data to a temporary directory to test delete commands against.
    """
    save_pipette_calibration(
        Point(1, 1, 1), "pip1", Mount.LEFT, "mytiprack", "opentrons/tip_rack/1"
    )
    save_pipette_calibration(
        Point(1, 1, 1), "pip2", Mount.RIGHT, "mytiprack", "opentrons/tip_rack/1"
    )


def test_delete_all_pipette_calibration(starting_calibration_data: Any) -> None:
    """
    Test delete all pipette calibrations.
    """
    assert get_pipette_offset("pip1", Mount.LEFT) is not None
    assert get_pipette_offset("pip2", Mount.RIGHT) is not None
    clear_pipette_offset_calibrations()
    assert get_pipette_offset("pip1", Mount.LEFT) is None
    assert get_pipette_offset("pip2", Mount.RIGHT) is None


def test_delete_specific_pipette_offset(starting_calibration_data: Any) -> None:
    """
    Test delete a specific pipette calibration.
    """
    assert get_pipette_offset("pip1", Mount.LEFT) is not None
    delete_pipette_offset_file("pip1", Mount.LEFT)
    assert get_pipette_offset("pip1", Mount.LEFT) is None


def test_save_pipette_calibration(
    ot_config_tempdir: Any, robot_model: "RobotModel"
) -> None:
    """
    Test saving pipette calibrations.
    """
    assert get_pipette_offset("pip1", Mount.LEFT) is None
    save_pipette_calibration(
        Point(1, 1, 1), "pip1", Mount.LEFT, "mytiprack", "opentrons/tip_rack/1"
    )
    save_pipette_calibration(
        Point(1, 1, 1), "pip2", Mount.RIGHT, "mytiprack", "opentrons/tip_rack/1"
    )
    assert get_pipette_offset("pip1", Mount.LEFT) is not None
    assert get_pipette_offset("pip2", Mount.RIGHT) is not None
    pip1 = get_pipette_offset("pip1", Mount.LEFT)
    assert pip1 is not None and pip1.offset == Point(1, 1, 1)
    pip2 = get_pipette_offset("pip1", Mount.LEFT)
    assert pip2 is not None and pip2.offset == Point(1, 1, 1)


def test_get_pipette_calibration(
    starting_calibration_data: Any, robot_model: "RobotModel"
) -> None:
    """
    Test ability to get a pipette calibration model.
    """
    from opentrons.calibration_storage.ot2.models.v1 import (
        InstrumentOffsetModel as OT2InstrModel,
    )

    pipette_data = get_pipette_offset("pip1", Mount.LEFT)
    assert pipette_data is not None

    assert pipette_data == OT2InstrModel(
        offset=Point(1, 1, 1),
        tiprack="mytiprack",
        uri="opentrons/tip_rack/1",
        last_modified=pipette_data.last_modified,
        source=cs_types.SourceType.user,
    )



"""Tests for LiquidClass properties and related functions."""

from typing import Union

import pytest

from opentrons_shared_data import load_shared_data
from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    Coordinate,
    LiquidClassSchemaV1,
    PositionReference,
)
from opentrons_shared_data.liquid_classes.types import TipPositionDict

from opentrons.protocol_api._liquid_properties import (
    LiquidHandlingPropertyByVolume,
    TipPosition,
    _ensure_validated_tip_position,
    build_aspirate_properties,
    build_multi_dispense_properties,
    build_single_dispense_properties,
)


def test_build_aspirate_settings() -> None:
    """It should convert the shared data aspirate settings to the PAPI type."""
    fixture_data = load_shared_data("liquid-class/fixtures/1/fixture_glycerol50.json")
    liquid_class_model = LiquidClassSchemaV1.model_validate_json(fixture_data)
    aspirate_data = liquid_class_model.byPipette[0].byTipType[0].aspirate

    aspirate_properties = build_aspirate_properties(aspirate_data)

    assert (
        aspirate_properties.submerge.start_position.position_reference.value
        == "liquid-meniscus"
    )
    assert aspirate_properties.submerge.start_position.offset == Coordinate(
        x=0, y=0, z=-5
    )
    assert aspirate_properties.submerge.speed == 100
    assert aspirate_properties.submerge.delay.enabled is True
    assert aspirate_properties.submerge.delay.duration == 1.5

    assert (
        aspirate_properties.retract.end_position.position_reference.value == "well-top"
    )
    assert aspirate_properties.retract.end_position.offset == Coordinate(x=0, y=0, z=5)
    assert aspirate_properties.retract.speed == 100
    assert aspirate_properties.retract.air_gap_by_volume.as_dict() == {
        5.0: 3.0,
        10.0: 4.0,
    }
    assert aspirate_properties.retract.touch_tip.enabled is True
    assert aspirate_properties.retract.touch_tip.z_offset == 2
    assert aspirate_properties.retract.touch_tip.mm_from_edge == 1
    assert aspirate_properties.retract.touch_tip.speed == 50
    assert aspirate_properties.retract.delay.enabled is True
    assert aspirate_properties.retract.delay.duration == 1

    assert (
        aspirate_properties.aspirate_position.position_reference.value == "well-bottom"
    )
    assert aspirate_properties.aspirate_position.offset == Coordinate(x=0, y=0, z=-5)
    assert aspirate_properties.flow_rate_by_volume.as_dict() == {10: 50.0}
    assert aspirate_properties.correction_by_volume.as_dict() == {
        1.0: -2.5,
        10.0: 3,
    }
    assert aspirate_properties.pre_wet is True
    assert aspirate_properties.mix.enabled is True
    assert aspirate_properties.mix.repetitions == 3
    assert aspirate_properties.mix.volume == 15
    assert aspirate_properties.delay.enabled is True
    assert aspirate_properties.delay.duration == 2
    assert aspirate_properties.as_shared_data_model() == aspirate_data


def test_aspirate_settings_overrides() -> None:
    """It should allow aspirate properties to be overridden with new values."""
    fixture_data = load_shared_data("liquid-class/fixtures/1/fixture_glycerol50.json")
    liquid_class_model = LiquidClassSchemaV1.model_validate_json(fixture_data)
    aspirate_data = liquid_class_model.byPipette[0].byTipType[0].aspirate

    aspirate_properties = build_aspirate_properties(aspirate_data)

    aspirate_properties.submerge.start_position.position_reference = "well-bottom"
    assert (
        aspirate_properties.submerge.start_position.position_reference.value
        == "well-bottom"
    )
    aspirate_properties.submerge.start_position.offset = 5, 0, 0
    assert aspirate_properties.submerge.start_position.offset == Coordinate(
        x=5, y=0, z=0
    )
    aspirate_properties.submerge.speed = 123
    assert aspirate_properties.submerge.speed == 123
    aspirate_properties.submerge.delay.enabled = False
    assert aspirate_properties.submerge.delay.enabled is False
    aspirate_properties.submerge.delay.duration = 5.1
    assert aspirate_properties.submerge.delay.duration == 5.1

    aspirate_properties.retract.end_position.position_reference = "well-center"
    assert (
        aspirate_properties.retract.end_position.position_reference.value
        == "well-center"
    )
    aspirate_properties.retract.end_position.offset = 0, 50, 0
    assert aspirate_properties.retract.end_position.offset == Coordinate(x=0, y=50, z=0)
    aspirate_properties.retract.speed = 987
    assert aspirate_properties.retract.speed == 987
    aspirate_properties.retract.touch_tip.enabled = False
    assert aspirate_properties.retract.touch_tip.enabled is False
    aspirate_properties.retract.touch_tip.z_offset = 2.34
    assert aspirate_properties.retract.touch_tip.z_offset == 2.34
    aspirate_properties.retract.touch_tip.mm_from_edge = 4.56
    assert aspirate_properties.retract.touch_tip.mm_from_edge == 4.56
    aspirate_properties.retract.touch_tip.speed = 501
    assert aspirate_properties.retract.touch_tip.speed == 501
    aspirate_properties.retract.delay.enabled = False
    assert aspirate_properties.retract.delay.enabled is False
    aspirate_properties.retract.delay.duration = 0.5
    assert aspirate_properties.retract.delay.duration == 0.5

    aspirate_properties.aspirate_position.position_reference = "liquid-meniscus"
    assert (
        aspirate_properties.aspirate_position.position_reference.value
        == "liquid-meniscus"
    )
    aspirate_properties.aspirate_position.offset = -1, -2, -3
    assert aspirate_properties.aspirate_position.offset == Coordinate(x=-1, y=-2, z=-3)
    aspirate_properties.pre_wet = False
    assert aspirate_properties.pre_wet is False
    aspirate_properties.mix.enabled = False
    assert aspirate_properties.mix.enabled is False
    aspirate_properties.mix.repetitions = 33
    assert aspirate_properties.mix.repetitions == 33
    aspirate_properties.mix.volume = 51
    assert aspirate_properties.mix.volume == 51
    aspirate_properties.delay.enabled = False
    assert aspirate_properties.delay.enabled is False
    aspirate_properties.delay.duration = 2.3
    assert aspirate_properties.delay.duration == 2.3


def test_build_single_dispense_settings() -> None:
    """It should convert the shared data single dispense settings to the PAPI type."""
    fixture_data = load_shared_data("liquid-class/fixtures/1/fixture_glycerol50.json")
    liquid_class_model = LiquidClassSchemaV1.model_validate_json(fixture_data)
    single_dispense_data = liquid_class_model.byPipette[0].byTipType[0].singleDispense

    single_dispense_properties = build_single_dispense_properties(single_dispense_data)

    assert (
        single_dispense_properties.submerge.start_position.position_reference.value
        == "liquid-meniscus"
    )
    assert single_dispense_properties.submerge.start_position.offset == Coordinate(
        x=0, y=0, z=-5
    )
    assert single_dispense_properties.submerge.speed == 100
    assert single_dispense_properties.submerge.delay.enabled is True
    assert single_dispense_properties.submerge.delay.duration == 1.5

    assert (
        single_dispense_properties.retract.end_position.position_reference.value
        == "well-top"
    )
    assert single_dispense_properties.retract.end_position.offset == Coordinate(
        x=0, y=0, z=5
    )
    assert single_dispense_properties.retract.speed == 100
    assert single_dispense_properties.retract.air_gap_by_volume.as_dict() == {
        5.0: 3.0,
        10.0: 4.0,
    }
    assert single_dispense_properties.retract.touch_tip.enabled is True
    assert single_dispense_properties.retract.touch_tip.z_offset == 2
    assert single_dispense_properties.retract.touch_tip.mm_from_edge == 1
    assert single_dispense_properties.retract.touch_tip.speed == 50
    assert single_dispense_properties.retract.blowout.enabled is True
    assert single_dispense_properties.retract.blowout.location is not None
    assert single_dispense_properties.retract.blowout.location.value == "trash"
    assert single_dispense_properties.retract.blowout.flow_rate == 100
    assert single_dispense_properties.retract.delay.enabled is True
    assert single_dispense_properties.retract.delay.duration == 1

    assert (
        single_dispense_properties.dispense_position.position_reference.value
        == "well-bottom"
    )
    assert single_dispense_properties.dispense_position.offset == Coordinate(
        x=0, y=0, z=-5
    )
    assert single_dispense_properties.flow_rate_by_volume.as_dict() == {
        10.0: 40.0,
        20.0: 30.0,
    }
    assert single_dispense_properties.correction_by_volume.as_dict() == {
        2.0: -1.5,
        20.0: 2,
    }
    assert single_dispense_properties.mix.enabled is True
    assert single_dispense_properties.mix.repetitions == 3
    assert single_dispense_properties.mix.volume == 15
    assert single_dispense_properties.push_out_by_volume.as_dict() == {
        10.0: 7.0,
        20.0: 10.0,
    }
    assert single_dispense_properties.delay.enabled is True
    assert single_dispense_properties.delay.duration == 2.5
    assert single_dispense_properties.as_shared_data_model() == single_dispense_data


def test_single_dispense_settings_override() -> None:
    """It should allow single dispense properties to be overridden with new values."""
    fixture_data = load_shared_data("liquid-class/fixtures/1/fixture_glycerol50.json")
    liquid_class_model = LiquidClassSchemaV1.model_validate_json(fixture_data)
    single_dispense_data = liquid_class_model.byPipette[0].byTipType[0].singleDispense

    single_dispense_properties = build_single_dispense_properties(single_dispense_data)

    single_dispense_properties.submerge.start_position.position_reference = (
        "well-bottom"
    )
    assert (
        single_dispense_properties.submerge.start_position.position_reference.value
        == "well-bottom"
    )
    single_dispense_properties.submerge.start_position.offset = 3, -2, 1
    assert single_dispense_properties.submerge.start_position.offset == Coordinate(
        x=3, y=-2, z=1
    )
    single_dispense_properties.submerge.speed = 111
    assert single_dispense_properties.submerge.speed == 111
    single_dispense_properties.submerge.delay.enabled = False
    assert single_dispense_properties.submerge.delay.enabled is False
    single_dispense_properties.submerge.delay.duration = 5.1
    assert single_dispense_properties.submerge.delay.duration == 5.1

    single_dispense_properties.retract.end_position.position_reference = "well-center"
    assert (
        single_dispense_properties.retract.end_position.position_reference.value
        == "well-center"
    )
    single_dispense_properties.retract.end_position.offset = -9, -8, -7
    assert single_dispense_properties.retract.end_position.offset == Coordinate(
        x=-9, y=-8, z=-7
    )
    single_dispense_properties.retract.speed = 222
    assert single_dispense_properties.retract.speed == 222
    single_dispense_properties.retract.touch_tip.enabled = False
    assert single_dispense_properties.retract.touch_tip.enabled is False
    single_dispense_properties.retract.touch_tip.z_offset = 2.34
    assert single_dispense_properties.retract.touch_tip.z_offset == 2.34
    single_dispense_properties.retract.touch_tip.mm_from_edge = 1.11
    assert single_dispense_properties.retract.touch_tip.mm_from_edge == 1.11
    single_dispense_properties.retract.touch_tip.speed = 543
    assert single_dispense_properties.retract.touch_tip.speed == 543
    single_dispense_properties.retract.blowout.enabled = False
    assert single_dispense_properties.retract.blowout.enabled is False
    single_dispense_properties.retract.blowout.location = "destination"
    assert single_dispense_properties.retract.blowout.location
    assert single_dispense_properties.retract.blowout.location.value == "destination"
    single_dispense_properties.retract.blowout.flow_rate = 3.21
    assert single_dispense_properties.retract.blowout.flow_rate == 3.21
    single_dispense_properties.retract.delay.enabled = False
    assert single_dispense_properties.retract.delay.enabled is False
    single_dispense_properties.retract.delay.duration = 0.1
    assert single_dispense_properties.retract.delay.duration == 0.1

    single_dispense_properties.retract.blowout.blowout_position = TipPosition(
        _position_reference=PositionReference.WELL_BOTTOM,
        _offset=Coordinate(x=10, y=20, z=30),
    )
    assert (
        single_dispense_properties.retract.blowout.blowout_position.position_reference
        == PositionReference.WELL_BOTTOM
    )
    assert (
        single_dispense_properties.retract.blowout.blowout_position.offset
        == Coordinate(x=10, y=20, z=30)
    )

    single_dispense_properties.retract.blowout.blowout_position = None
    assert single_dispense_properties.retract.blowout.blowout_position is None
    single_dispense_properties.dispense_position = (
        single_dispense_properties.submerge.start_position
    )
    assert (
        single_dispense_properties.dispense_position.position_reference.value
        == "well-bottom"
    )
    assert single_dispense_properties.dispense_position.offset == Coordinate(
        x=3, y=-2, z=1
    )
    single_dispense_properties.mix.enabled = False
    assert single_dispense_properties.mix.enabled is False
    single_dispense_properties.mix.repetitions = 15
    assert single_dispense_properties.mix.repetitions == 15
    single_dispense_properties.mix.volume = 3
    assert single_dispense_properties.mix.volume == 3
    single_dispense_properties.delay.enabled = False
    assert single_dispense_properties.delay.enabled is False
    single_dispense_properties.delay.duration = 25.25
    assert single_dispense_properties.delay.duration == 25.25


def test_build_multi_dispense_settings() -> None:
    """It should convert the shared data multi dispense settings to the PAPI type."""
    fixture_data = load_shared_data("liquid-class/fixtures/1/fixture_glycerol50.json")
    liquid_class_model = LiquidClassSchemaV1.model_validate_json(fixture_data)
    multi_dispense_data = liquid_class_model.byPipette[0].byTipType[0].multiDispense

    assert multi_dispense_data is not None
    multi_dispense_properties = build_multi_dispense_properties(multi_dispense_data)
    assert multi_dispense_properties is not None

    assert (
        multi_dispense_properties.submerge.start_position.position_reference.value
        == "liquid-meniscus"
    )
    assert multi_dispense_properties.submerge.start_position.offset == Coordinate(
        x=0, y=0, z=-5
    )
    assert multi_dispense_properties.submerge.speed == 100
    assert multi_dispense_properties.submerge.delay.enabled is True
    assert multi_dispense_properties.submerge.delay.duration == 1.5

    assert (
        multi_dispense_properties.retract.end_position.position_reference.value
        == "well-top"
    )
    assert multi_dispense_properties.retract.end_position.offset == Coordinate(
        x=0, y=0, z=5
    )
    assert multi_dispense_properties.retract.speed == 100
    assert multi_dispense_properties.retract.air_gap_by_volume.as_dict() == {
        5.0: 3.0,
        10.0: 4.0,
    }
    assert multi_dispense_properties.retract.touch_tip.enabled is True
    assert multi_dispense_properties.retract.touch_tip.z_offset == 2
    assert multi_dispense_properties.retract.touch_tip.mm_from_edge == 1
    assert multi_dispense_properties.retract.touch_tip.speed == 50
    assert multi_dispense_properties.retract.blowout.enabled is False
    assert multi_dispense_properties.retract.blowout.location is None
    assert multi_dispense_properties.retract.blowout.flow_rate is None
    assert multi_dispense_properties.retract.delay.enabled is True
    assert multi_dispense_properties.retract.delay.duration == 1

    assert (
        multi_dispense_properties.dispense_position.position_reference.value
        == "well-bottom"
    )
    assert multi_dispense_properties.dispense_position.offset == Coordinate(
        x=0, y=0, z=-5
    )
    assert multi_dispense_properties.flow_rate_by_volume.as_dict() == {
        10.0: 40.0,
        20.0: 30.0,
    }
    assert multi_dispense_properties.correction_by_volume.as_dict() == {
        3.0: -0.5,
        30.0: 1,
    }
    assert multi_dispense_properties.conditioning_by_volume.as_dict() == {
        5.0: 5.0,
    }
    assert multi_dispense_properties.disposal_by_volume.as_dict() == {
        5.0: 3.0,
    }
    assert multi_dispense_properties.delay.enabled is True
    assert multi_dispense_properties.delay.duration == 1
    assert multi_dispense_properties.as_shared_data_model() == multi_dispense_data


def test_multi_dispense_settings_override() -> None:
    """It should allow multi dispense properties to be overridden with new values."""
    fixture_data = load_shared_data("liquid-class/fixtures/1/fixture_glycerol50.json")
    liquid_class_model = LiquidClassSchemaV1.model_validate_json(fixture_data)
    multi_dispense_data = liquid_class_model.byPipette[0].byTipType[0].multiDispense
    assert multi_dispense_data is not None
    multi_dispense_properties = build_multi_dispense_properties(multi_dispense_data)
    assert multi_dispense_properties is not None

    multi_dispense_properties.submerge.start_position.position_reference = "well-bottom"
    assert (
        multi_dispense_properties.submerge.start_position.position_reference.value
        == "well-bottom"
    )
    multi_dispense_properties.submerge.start_position.offset = 3, -2, 1
    assert multi_dispense_properties.submerge.start_position.offset == Coordinate(
        x=3, y=-2, z=1
    )
    multi_dispense_properties.submerge.speed = 111
    assert multi_dispense_properties.submerge.speed == 111
    multi_dispense_properties.submerge.delay.enabled = False
    assert multi_dispense_properties.submerge.delay.enabled is False
    multi_dispense_properties.submerge.delay.duration = 5.1
    assert multi_dispense_properties.submerge.delay.duration == 5.1

    multi_dispense_properties.retract.end_position.position_reference = "well-center"
    assert (
        multi_dispense_properties.retract.end_position.position_reference.value
        == "well-center"
    )
    multi_dispense_properties.retract.end_position.offset = -9, -8, -7
    assert multi_dispense_properties.retract.end_position.offset == Coordinate(
        x=-9, y=-8, z=-7
    )
    multi_dispense_properties.retract.speed = 222
    assert multi_dispense_properties.retract.speed == 222
    multi_dispense_properties.retract.touch_tip.enabled = False
    assert multi_dispense_properties.retract.touch_tip.enabled is False
    multi_dispense_properties.retract.touch_tip.z_offset = 2.34
    assert multi_dispense_properties.retract.touch_tip.z_offset == 2.34
    multi_dispense_properties.retract.touch_tip.mm_from_edge = 1.11
    assert multi_dispense_properties.retract.touch_tip.mm_from_edge == 1.11
    multi_dispense_properties.retract.touch_tip.speed = 543
    assert multi_dispense_properties.retract.touch_tip.speed == 543
    multi_dispense_properties.retract.blowout.enabled = False
    assert multi_dispense_properties.retract.blowout.enabled is False
    multi_dispense_properties.retract.blowout.location = "destination"
    assert multi_dispense_properties.retract.blowout.location
    assert multi_dispense_properties.retract.blowout.location.value == "destination"
    multi_dispense_properties.retract.blowout.flow_rate = 3.21
    assert multi_dispense_properties.retract.blowout.flow_rate == 3.21
    multi_dispense_properties.retract.delay.enabled = False
    assert multi_dispense_properties.retract.delay.enabled is False
    multi_dispense_properties.retract.delay.duration = 0.1
    assert multi_dispense_properties.retract.delay.duration == 0.1

    multi_dispense_properties.dispense_position.position_reference = "liquid-meniscus"
    assert (
        multi_dispense_properties.dispense_position.position_reference.value
        == "liquid-meniscus"
    )
    multi_dispense_properties.dispense_position.offset = 11, 22, -33
    assert multi_dispense_properties.dispense_position.offset == Coordinate(
        x=11, y=22, z=-33
    )
    multi_dispense_properties.delay.enabled = False
    assert multi_dispense_properties.delay.enabled is False
    multi_dispense_properties.delay.duration = 25.25
    assert multi_dispense_properties.delay.duration == 25.25


def test_build_multi_dispense_settings_none(
    minimal_liquid_class_def2: LiquidClassSchemaV1,
) -> None:
    """It should return None if there are no multi dispense properties in the model."""
    transfer_settings = minimal_liquid_class_def2.byPipette[0].byTipType[0]
    assert build_multi_dispense_properties(transfer_settings.multiDispense) is None


def test_liquid_handling_property_by_volume() -> None:
    """It should create a class that can interpolate values and add and delete new points."""
    subject = LiquidHandlingPropertyByVolume([(5.0, 50.0), (10.0, 250.0)])
    assert subject.as_dict() == {5.0: 50, 10.0: 250}
    assert subject.get_for_volume(7) == 130.0
    assert subject.as_list_of_tuples() == [(5.0, 50.0), (10.0, 250.0)]

    subject.set_for_volume(volume=7, value=175.5)
    assert subject.as_dict() == {
        5.0: 50,
        10.0: 250,
        7.0: 175.5,
    }
    assert subject.get_for_volume(7) == 175.5

    subject.delete_for_volume(7)
    assert subject.as_dict() == {5.0: 50, 10.0: 250}
    assert subject.get_for_volume(7) == 130.0

    with pytest.raises(KeyError, match="No value set for volume"):
        subject.delete_for_volume(7)

    # Test bounds
    assert subject.get_for_volume(1) == 50.0
    assert subject.get_for_volume(1000) == 250.0

    # Test fixed overrides
    subject.set_for_all_volumes(4321)
    for volume in (0, 1, 7, 100):
        assert subject.get_for_volume(volume) == 4321

    # Test resetting to default
    subject.reset_values()
    assert subject.as_dict() == {5.0: 50, 10.0: 250}
    assert subject.get_for_volume(7) == 130.0

    # Test clearing all values
    subject.clear_values()
    assert subject.as_dict() == {}
    with pytest.raises(ValueError, match="No properties found for any volumes"):
        subject.get_for_volume(7)


def test_non_existent_property_raises_error() -> None:
    """It should raise an attribute error if the set property does not exist."""
    fixture_data = load_shared_data("liquid-class/fixtures/1/fixture_glycerol50.json")
    liquid_class_model = LiquidClassSchemaV1.model_validate_json(fixture_data)
    aspirate_data = liquid_class_model.byPipette[0].byTipType[0].aspirate

    aspirate_properties = build_aspirate_properties(aspirate_data)

    with pytest.raises(AttributeError):
        aspirate_properties.mix.enable = True  # type: ignore[attr-defined]


@pytest.mark.parametrize(
    "tip_position, expected_position",
    [
        (
            TipPosition(
                _position_reference=PositionReference.WELL_BOTTOM,
                _offset=Coordinate(x=1, y=2, z=3),
            ),
            TipPosition(
                _position_reference=PositionReference.WELL_BOTTOM,
                _offset=Coordinate(x=1, y=2, z=3),
            ),
        ),
        (
            {"position_reference": "well-top", "offset": {"x": 10, "y": 20, "z": 30}},
            TipPosition(
                _position_reference=PositionReference.WELL_TOP,
                _offset=Coordinate(x=10, y=20, z=30),
            ),
        ),
    ],
)
def test_ensure_validated_tip_position(
    tip_position: Union[TipPosition, TipPositionDict], expected_position: TipPosition
) -> None:
    """It should set tip position properties correctly."""
    validated_position = _ensure_validated_tip_position(tip_position)
    assert validated_position == expected_position


def test_ensure_validated_tip_position_raises_error() -> None:
    """It should raise an error if the tip position is invalid."""
    with pytest.raises(TypeError):
        _ensure_validated_tip_position("invalid_tip_position")  # type: ignore[arg-type]

    with pytest.raises(
        ValueError,
        match="Input should be 'well-bottom', 'well-top', 'well-center' or 'liquid-meniscus'",
    ):
        _ensure_validated_tip_position(
            {"position_reference": "invalid", "offset": {"x": 10, "y": 20, "z": 30}}  # type: ignore[arg-type]
        )

    with pytest.raises(ValueError, match="Extra inputs are not permitted"):
        _ensure_validated_tip_position(
            {  # type: ignore[arg-type]
                "position_reference": "well-bottom",
                "offset": {"x": 10, "y": 20, "z": 30, "w": 3},
            }
        )



"""Test calibrate-pipette command."""

from __future__ import annotations

import inspect
from typing import TYPE_CHECKING

import pytest
from decoy import Decoy

from opentrons_shared_data.errors.exceptions import EarlyCapacitiveSenseTrigger

from opentrons.hardware_control import ot3_calibration as calibration
from opentrons.hardware_control.api import API
from opentrons.hardware_control.types import OT3Mount
from opentrons.protocol_engine.commands.calibration.calibrate_pipette import (
    CalibratePipetteImplementation,
    CalibratePipetteParams,
    CalibratePipetteResult,
)
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.errors.exceptions import HardwareNotSupportedError
from opentrons.protocol_engine.types import InstrumentOffsetVector
from opentrons.types import MountType, Point

if TYPE_CHECKING:
    from opentrons.hardware_control.ot3api import OT3API


@pytest.fixture(autouse=True)
def _mock_ot3_calibration(decoy: Decoy, monkeypatch: pytest.MonkeyPatch) -> None:
    for name, func in inspect.getmembers(calibration, inspect.isfunction):
        monkeypatch.setattr(calibration, name, decoy.mock(func=func))


@pytest.mark.ot3_only
async def test_calibrate_pipette_implementation(
    decoy: Decoy, ot3_hardware_api: OT3API
) -> None:
    """Test Calibration command execution."""
    subject = CalibratePipetteImplementation(hardware_api=ot3_hardware_api)

    params = CalibratePipetteParams(
        mount=MountType.LEFT,
    )

    decoy.when(
        await calibration.find_pipette_offset(
            hcapi=ot3_hardware_api, mount=OT3Mount.LEFT, slot=5
        )
    ).then_return(Point(x=3, y=4, z=6))

    result = await subject.execute(params)

    decoy.verify(
        await ot3_hardware_api.save_instrument_offset(
            mount=OT3Mount.LEFT, delta=Point(x=3, y=4, z=6)
        ),
        times=1,
    )

    assert result == SuccessData(
        public=CalibratePipetteResult(
            pipetteOffset=InstrumentOffsetVector(x=3, y=4, z=6)
        ),
    )


@pytest.mark.ot3_only
async def test_calibrate_pipette_does_not_save_during_error(
    decoy: Decoy, ot3_hardware_api: OT3API
) -> None:
    """Data should not be saved when an error is raised."""
    subject = CalibratePipetteImplementation(hardware_api=ot3_hardware_api)

    params = CalibratePipetteParams(
        mount=MountType.LEFT,
    )

    decoy.when(
        await calibration.find_pipette_offset(
            hcapi=ot3_hardware_api, mount=OT3Mount.LEFT, slot=5
        )
    ).then_raise(EarlyCapacitiveSenseTrigger(5.0, 3.0))

    with pytest.raises(EarlyCapacitiveSenseTrigger):
        await subject.execute(params)

    decoy.verify(
        await ot3_hardware_api.save_instrument_offset(
            mount=OT3Mount.LEFT, delta=Point(x=3, y=4, z=6)
        ),
        times=0,
    )


@pytest.mark.ot3_only
async def test_calibrate_pipette_implementation_wrong_hardware(
    decoy: Decoy, ot2_hardware_api: API
) -> None:
    """Should raise an unsupported hardware error."""
    subject = CalibratePipetteImplementation(hardware_api=ot2_hardware_api)

    params = CalibratePipetteParams(
        mount=MountType.LEFT,
    )

    with pytest.raises(HardwareNotSupportedError):
        await subject.execute(params)



import typing

import pytest

from opentrons.calibration_storage import (
    types as cs_types,
)
from opentrons.calibration_storage.ot3 import gripper_offset as gripper
from opentrons.calibration_storage.ot3 import models
from opentrons.types import Point


@pytest.fixture
def starting_calibration_data(
    ot_config_tempdir: typing.Any, enable_ot3_hardware_controller: typing.Any
) -> None:
    """
    Starting calibration data fixture.

    Adds dummy data to a temporary directory to test delete commands against.
    """
    gripper.save_gripper_calibration(Point(1, 1, 1), "gripper1")
    gripper.save_gripper_calibration(Point(1, 2, 1), "gripper2")
    gripper.save_gripper_jaw_width_data(15.5, "gripper1")
    gripper.save_gripper_jaw_width_data(15.5, "gripper2")


def test_delete_all_gripper_calibration(starting_calibration_data: typing.Any) -> None:
    """
    Test delete all gripper calibrations.
    """
    assert gripper.get_gripper_calibration_offset("gripper1") is not None
    assert gripper.get_gripper_calibration_offset("gripper2") is not None
    assert gripper.get_gripper_jaw_width_data("gripper1") is not None
    assert gripper.get_gripper_jaw_width_data("gripper2") is not None
    gripper.clear_gripper_calibration_offsets()
    gripper.clear_gripper_jaw_width_data()
    assert gripper.get_gripper_calibration_offset("gripper1") is None
    assert gripper.get_gripper_calibration_offset("gripper2") is None
    assert gripper.get_gripper_jaw_width_data("gripper1") is None
    assert gripper.get_gripper_jaw_width_data("gripper2") is None


def test_delete_gripper_calibration(starting_calibration_data: typing.Any) -> None:
    """
    Test delete a single gripper calibration.
    """
    assert gripper.get_gripper_calibration_offset("gripper1") is not None
    gripper.delete_gripper_calibration_file("gripper1")
    assert gripper.get_gripper_calibration_offset("gripper1") is None


def test_save_gripper_calibration(
    ot_config_tempdir: typing.Any, enable_ot3_hardware_controller: typing.Any
) -> None:
    """
    Test saving gripper calibrations.
    """
    assert gripper.get_gripper_calibration_offset("gripper1") is None
    gripper.save_gripper_calibration(Point(1, 1, 1), "gripper1")
    gripper_offset = gripper.get_gripper_calibration_offset("gripper1")
    assert gripper_offset is not None
    assert gripper_offset.offset == Point(1, 1, 1)


def test_save_gripper_jaw_width(
    ot_config_tempdir: typing.Any, enable_ot3_hardware_controller: typing.Any
) -> None:
    """
    Test saving gripper jaw calibrations.
    """
    assert gripper.get_gripper_jaw_width_data("gripper1") is None
    gripper.save_gripper_jaw_width_data(15.44, "gripper1")
    gripper_offset = gripper.get_gripper_jaw_width_data("gripper1")
    assert gripper_offset is not None
    assert gripper_offset.encoderPositionAtJawClosed == 15.44


def test_get_gripper_calibration(
    starting_calibration_data: typing.Any, enable_ot3_hardware_controller: typing.Any
) -> None:
    """
    Test ability to get a gripper calibration schema.
    """
    gripper_data = gripper.get_gripper_calibration_offset("gripper1")
    assert gripper_data is not None
    assert gripper_data == models.v1.InstrumentOffsetModel(
        offset=Point(1, 1, 1),
        lastModified=gripper_data.lastModified,
        source=cs_types.SourceType.user,
    )


def test_get_gripper_jaw_width(
    starting_calibration_data: typing.Any, enable_ot3_hardware_controller: typing.Any
) -> None:
    """
    Test ability to get a gripper jaw calibration schema.
    """
    gripper_data = gripper.get_gripper_jaw_width_data("gripper1")
    assert gripper_data is not None
    assert gripper_data == models.v1.GripperJawWidthModel(
        encoderPositionAtJawClosed=15.5,
        lastModified=gripper_data.lastModified,
        source=cs_types.SourceType.user,
    )
